#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# This code is used check job generation
#
import os
import shutil   # used to remove directory tree

import ecflow_test_util as Test
from ecflow import Defs, JobCreationCtrl, TaskVec, File, Client, debug_build

def ecf_includes() : return File.source_dir() + "/libs/pyext/test/data/includes"

def suite_name(): return "suite_job_gen_" + str(os.getpid())  # allow python2 and python3 to run in parallel

def create_defs(ecf_home,task_vec):
    defs = Defs();
    suite = defs.add_suite( suite_name() );
    suite.add_variable("ECF_HOME",ecf_home);
    suite.add_variable("ECF_CLIENT_EXE_PATH",File.find_client());
    suite.add_variable("SLEEPTIME","1");
    suite.add_variable("ECF_INCLUDE",ecf_includes());
    
    fam =  suite.add_family("family")
    t1 = fam.add_task("t1")
    t1.add_event("eventname");
    t1.add_meter("meter",0,100,100)
    
    t2 = fam.add_task("t2")
    t3 = fam.add_task("t3")
    t4 = fam.add_task("dummy_task")
    t4.add_variable("ECF_DUMMY_TASK","any"); # stop job generation errors for tasks without an .ecf
    task_vec.append(t1);
    task_vec.append(t2);
    task_vec.append(t3);
    task_vec.append(t4);
    
    # Create dir. **NOTE**, if the second component of os.path.joind(..) is absolute, 
    # the first path is ignored, hence remove first char
    dir_to_create = os.path.join(str(ecf_home),fam.get_abs_node_path()[1:])
    print("Creating dir ",dir_to_create)
    os.makedirs( dir_to_create  )
     
    # copy ecf files from ecflow source, to create unique file per process
    src_t1  = os.path.join(ecf_home, "suite_job_gen" , "family" , "t1.ecf")
    dest_t1 = os.path.join(ecf_home,  t1.get_abs_node_path()[1:] + ".ecf" )
    src_t2  = os.path.join(ecf_home, "suite_job_gen" , "family" , "t2.ecf")
    dest_t2 = os.path.join(ecf_home,  t2.get_abs_node_path()[1:] + ".ecf" )
    src_t3  = os.path.join(ecf_home, "suite_job_gen" , "family" , "t3.ecf")
    dest_t3 = os.path.join(ecf_home,  t3.get_abs_node_path()[1:] + ".ecf" )
 
    shutil.copy(src_t1, dest_t1)
    shutil.copy(src_t2, dest_t2)
    shutil.copy(src_t3, dest_t3)
    
    return defs


def delete_jobs(task_vec, ecf_home):
    print("delete jobs")
    for task in task_vec:
        the_job_file = ecf_home + task.get_abs_node_path() + ".job" + task.get_try_no()
        if os.path.exists(the_job_file) :
            print("removing file " + the_job_file)
            try: os.remove(the_job_file)
            except: pass
        man_file = ecf_home + task.get_abs_node_path() + ".man"  
        if os.path.exists(man_file) :
            print("removing man_file " + man_file)
            try: os.remove(man_file)
            except: pass
   
def check_jobs(task_vec, ecf_home):
    print("Check job file exists")
    for task in task_vec:
        variable = task.find_variable("ECF_DUMMY_TASK")
        if not variable.empty(): continue;
        the_job_file = ecf_home + task.get_abs_node_path() + ".job" + task.get_try_no()
      
        if os.path.exists(the_job_file) :
            print("Found job file " + the_job_file)
        else:
            assert False, "Could not find job file " + the_job_file


if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
    
    workspace = File.source_dir();
    print(workspace)

    ecf_home = workspace + "/libs/pyext/test/data/ECF_HOME"
    task_vec = TaskVec()
    defs = create_defs(ecf_home,task_vec)
    print(str(defs))

    print("Generate jobs for *ALL* tasks, to default locations ECF_HOME/ECF_NAME.job0") 
    print(defs.check_job_creation())   
    check_jobs(task_vec,ecf_home)
    delete_jobs(task_vec,ecf_home)
       
    print("\nGenerate jobs for *ALL* tasks, to default locations ECF_HOME/ECF_NAME.job0")
    job_ctrl = JobCreationCtrl()
    defs.check_job_creation( job_ctrl )       
    print(job_ctrl.get_error_msg())
    check_jobs(task_vec,ecf_home)
    delete_jobs(task_vec,ecf_home)
    
    print("\nGenerate jobs for all nodes, under path, to default locations ECF_HOME/ECF_NAME.job0")    
    job_ctrl = JobCreationCtrl()
    job_ctrl.set_node_path( task_vec[0].get_abs_node_path() )    
    defs.check_job_creation(job_ctrl)       
    print(job_ctrl.get_error_msg());
    delete_jobs(task_vec,ecf_home)
   
    print("\nGenerate jobs for all tasks, to the specified directory")
    # Directory will automatically created under the provided directory
    job_ctrl = JobCreationCtrl()
    job_ctrl.set_dir_for_job_creation(workspace + "/libs/pyext/test/data")  # generate jobs file under this directory
    defs.check_job_creation(job_ctrl)
    print(job_ctrl.get_error_msg())
    
    generated_dir = job_ctrl.get_dir_for_job_creation() + "/" + suite_name()
    print("removing directory tree " + generated_dir)
    shutil.rmtree(generated_dir)      
 
    # Create jobs for all task, to a TMP directory ($TMPDIR/ecf_check_job_creation/ECF_NAME.job0
    # When run via rsh then TMPDIR may not be defined, hence expect exception to be thrown
    try:
        job_ctrl = JobCreationCtrl()                    
        job_ctrl.generate_temp_dir()       
        defs.check_job_creation(job_ctrl)
        print(job_ctrl.get_error_msg())
        print("removing directory tree " + job_ctrl.get_dir_for_job_creation())
        shutil.rmtree(job_ctrl.get_dir_for_job_creation(),ignore_errors=True)     
        delete_jobs(task_vec,ecf_home)
    except RuntimeError as e:
        print("failed: " + str(e))  
        
    try:
        dir_to_del = os.path.join(ecf_home,suite_name())
        print("removing directory tree " + dir_to_del )
        shutil.rmtree(dir_to_del)     
        print("All test pass")
    except RuntimeError as e:
        print("failed: " + str(e))    



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import datetime
import os
import sys

# Adds path to the folder _ext, where extensions are stored
sys.path.insert(0, os.path.abspath("."))
sys.path.append(os.path.abspath("./_ext"))

# build_path = "/Users/cgr/build/ecflow/debug/Pyext/python3"
# sys.path.insert(0, build_path)

# -- Project information -----------------------------------------------------

project = "ecflow"
author = "ECMWF"

year = datetime.datetime.now().year
years = "2009-%s" % (year,)
copyright = "%s, European Centre for Medium-Range Weather Forecasts (ECMWF)" % (years,)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx_rtd_theme"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "python_api_old",
    "build_python_api",
]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["css/custom_style.css"]

html_logo = ""

highlight_language = "none"

numfig = True

rst_prolog = """
.. role:: mval
"""

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#


def get_ecflow_version():
    version = "5.13.8"
    ecflow_version = version.split(".")
    print("Extracted ecflow version '" + str(ecflow_version))
    return ecflow_version


# import sphinx.ext.autodoc

# _RST_FP = open("MY_PY.rst", "w")


# def add_line(self, line, source, *lineno):
#     """Append one line of generated reST to the output."""
#     _RST_FP.write(line + "\n")
#     # print(f"source={source}")
#     self.directive.result.append(self.indent + line, source, *lineno)


# sphinx.ext.autodoc.Documenter.add_line = add_line

# The full version, including alpha/beta/rc tags.
ecflow_version_list = get_ecflow_version()
assert len(ecflow_version_list) == 3, "Expected version to have release, major,minor"
version = ".".join(ecflow_version_list[:2])
release = ".".join(ecflow_version_list)



import os
import ecflow


class DefaultVariables(object):
    """Provide the setup variables for each suite"""

    def add_to(self, node):
        """Adds ECF_INCLUDE,ECF_HOME to the input node"""
        node.add_variable("ECF_INCLUDE", os.path.join(os.getenv("HOME"), "course"))
        node.add_variable("ECF_HOME", os.path.join(os.getenv("HOME"), "course"))


class BaseSuiteBuilder(object):
    """Abstract class. Add default variables to suite and enable job creation
    checking for any derived suite
    """

    def __init__(self, default_variables):
        self.defs = ecflow.Defs()
        # use derived class name as suite name
        self.suite = self.defs.add_suite(type(self).__name__)
        default_variables.add_to(self.suite)

    def _build_definition_hook(self):
        """Derived suite should override this function to build the suite
        Should not be called explicitly. How could we enforce this ?
        """
        pass

    def setup(self):
        """Template/skeleton function.
        Provides common algorithm for *all* derivatives
        Uses Hollywood principle.
        """

        # Build the suite
        self._build_definition_hook()

        # check job creation. Could use an assert
        job_check_result = self.defs.check_job_creation()
        if len(job_check_result) != 0:
            print("Job creation failed\n" + job_check_result)

        # Check trigger expressions and limit references
        print(self.defs.check())

        # Allows definition creation to be separated from the load
        # Use the class name as the name of the definition file
        self.defs.save_as_defs(type(self).__name__ + ".def")


class SuiteBuilder(BaseSuiteBuilder):
    """My example suite. Generates SuiteBuilder.def"""

    def __init__(self, default_variables):
        BaseSuiteBuilder.__init__(self, default_variables)

    def _build_definition_hook(self):
        f1 = self.suite.add_family("family")
        f1.add_task("task1")
        f1.add_task("task2")


class TestSuite(BaseSuiteBuilder):
    """My test suite. Generates TestSuite.def"""

    def __init__(self, default_variables):
        BaseSuiteBuilder.__init__(self, default_variables)

    def _build_definition_hook(self):
        self.suite.add_task("task1")


if __name__ == "__main__":

    my_suite = SuiteBuilder(DefaultVariables())
    my_suite.setup()

    my_test_suite = TestSuite(DefaultVariables())
    my_test_suite.setup()



#!/usr/bin/env python3

import glob
import os
import re

import yaml


def fix_glossary():
    with open("glossary.yaml", "r") as f:
        glossary_items = yaml.load(f, Loader=yaml.FullLoader)

    for fname in glob.glob("rst/*.rst"):
        if os.path.basename(fname)[0].isupper():
            t = ""
            with open(fname, "r") as f:
                t = f.read()
                if t:
                    for item in glossary_items:
                        t = re.sub(rf"`{item}`_", rf":term:`{item}`", t)
                    for item in ['cron definition<text_based_def_cron>']:
                        t = re.sub(rf"`{item}`_", rf":ref:`{item}`", t)

            if t:
                with open(fname, "w") as f:
                    f.write(t)


fix_glossary()



# Copyright 2009-2019 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import time
import argparse  # for argument parsing
from ecflow import Client, debug_build


def timing_decorator(function_to_time):
    def scafold(*function_args):
        # function_args is a tuple, the second argument is the suite name
        start = time.time()
        function_return = function_to_time(*function_args)
        end = time.time()
        print(
            "%s function took %0.3f ms for %s"
            % (function_to_time.func_name, (end - start) * 1000.0, function_args[1])
        )
        return function_return

    return scafold


@timing_decorator
def sync_local(ci, suite_name):
    ci.sync_local()


if __name__ == "__main__":

    DESC = """This test is use to show the performance of sync local on the given server
              It should show that it is quicker to register and then sync_local()
              as this limits the amount of data that needs to be down loaded from the  server.
              Usage:
                   TestSyncLocalPerf.py --host <hostname> --port <portname> 
            """
    PARSER = argparse.ArgumentParser(
        description=DESC, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    PARSER.add_argument(
        "--host",
        default="localhost",
        help="The name of the host machine, defaults to 'localhost'",
    )
    PARSER.add_argument(
        "--port", default="3141", help="The port on the host, defaults to 3141"
    )
    ARGS = PARSER.parse_args()
    print(ARGS)

    print("####################################################################")
    print(
        "Test performance of sync local using "
        + Client().version()
        + " debug build("
        + str(debug_build())
        + ")"
    )
    print("####################################################################")

    CL = Client(ARGS.host, ARGS.port)

    sync_local(CL, "* All suites *")  # timing for downloading *all* suites
    print("")

    suites = CL.get_defs().suites
    for suite in suites:
        CL.ch_register(
            False, [suite.name()]
        )  # register interest in single suite, corresponding handle is attached to CL
        sync_local(CL, suite.name())  # timing to download a single suite
        CL.ch_drop()  # drop the last registered handle on CL



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# Simple check for get all tasks
from ecflow import Defs, Client, debug_build
import ecflow_test_util as Test
import os

def create_defs():
    defs = Defs()
    suite = defs.add_suite("test_get_all")
    suite.add_task("t0")
    fam = suite.add_family("f1")
    fam.add_task("t1") 
    fam.add_task("t2") 
    fam2 = fam.add_family("f2")
    fam2.add_task("t3") 
    return defs
   
def test_get_all_tasks(defs):
  
    task_vec2 = defs.get_all_tasks()
    assert len(task_vec2) == 4, "Expected four tasks, but found " + str(len(task_vec2))
   
    assert task_vec2[0].name() == "t0", "Expected task of name t0 but found " + task_vec2[0].name()
    assert task_vec2[1].name() == "t1", "Expected task of name t1 but found " + task_vec2[1].name()
    assert task_vec2[2].name() == "t2", "Expected task of name t2 but found " + task_vec2[2].name()
    assert task_vec2[3].name() == "t3", "Expected task of name t3 but found " + task_vec2[3].name()
   
    print("test_get_all_tasks PASSED")
   
def test_get_all_nodes(defs):
    
    # Get all the nodes make sure they are in order
    node_vec = defs.get_all_nodes()
    assert len(node_vec) == 7, "Expected seven nodes, but found " + str(len(node_vec))
    assert node_vec[0].name() == "test_get_all", "Nodes should be returned in order, expected test_get_all but found" + node_vec[0].name()
    assert node_vec[1].name() == "t0", "Nodes should be returned in order, expected 't0' but found" + node_vec[1].name()
    assert node_vec[2].name() == "f1", "Nodes should be returned in order, expected 'f1' but found" + node_vec[2].name()
    assert node_vec[3].name() == "t1", "Nodes should be returned in order, expected 't1' but found" + node_vec[3].name()
    assert node_vec[4].name() == "t2", "Nodes should be returned in order, expected 't2' but found" + node_vec[4].name()
    assert node_vec[5].name() == "f2", "Nodes should be returned in order, expected 'f2' but found" + node_vec[5].name()
    assert node_vec[6].name() == "t3", "Nodes should be returned in order, expected 't3' but found" + node_vec[6].name()

    # test empty defs
    empty_defs = Defs()
    node_vec = empty_defs.get_all_nodes()
    assert len(node_vec) == 0, "Expected zero nodes, for an empty defs but found " + str(len(node_vec))

    # test one suite defs
    one_suite_defs = Defs()
    one_suite_defs.add_suite("s1")
    node_vec = one_suite_defs.get_all_nodes()
    assert len(node_vec) == 1, "Expected one node, but found " + str(len(node_vec))

    print("test_get_all_nodes PASSED")
    
def test_get_all_nodes_from_nodes():
    defs = Defs()
    suite = defs.add_suite("test_get_all")
    node_vec = suite.get_all_nodes()
    assert len(node_vec) == 1, "Expected one nodes but found " + str(len(node_vec))

    fam1 = suite.add_family("f1")
    fam2 = suite.add_family("f2")
    fam3 = suite.add_family("f3")
    node_vec = suite.get_all_nodes()
    assert len(node_vec) == 4, "Expected 4 nodes but found " + str(len(node_vec))

    node_vec = fam1.get_all_nodes()
    assert len(node_vec) == 1, "Expected 1 nodes but found " + str(len(node_vec))

    f1_t1 = fam1.add_task("t1") 
    f1_t2 = fam1.add_task("t2") 
    node_vec = fam1.get_all_nodes()
    assert len(node_vec) == 3, "Expected 3 nodes but found " + str(len(node_vec))

    node_vec = f1_t1.get_all_nodes()
    assert len(node_vec) == 1, "Expected 1 nodes but found " + str(len(node_vec))

    node_vec = suite.get_all_nodes()
    assert len(node_vec) == 6, "Expected 6 nodes but found " + str(len(node_vec))

    print("test_get_all_nodes_from_nodes PASSED")
   

if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
 
    test_get_all_tasks(create_defs())
    test_get_all_nodes(create_defs())
    test_get_all_nodes_from_nodes()
    print("All Tests pass")    


import ecflow

try:
    print("Loading definition in 'test.def' into the server")
    ci = ecflow.Client()
    ci.load("test.def")  # read definition from disk and load into the server
except RuntimeError as e:
    print("Failed:", e)



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

#  code for testing pointers and hierarchy in python

from ecflow import Suite, Family, Task, Defs, Client, debug_build, Edit
import ecflow_test_util as Test
import os

def create_defs(name=""):
    suite_name = name
    if len(suite_name) == 0: suite_name = "s1"
    defs =  Defs(Edit(FRED_2="XX",DEFS_VAR="0"),
                Suite(suite_name, Edit(SUITE_VAR="1",FRED_1="%FRED_2%"),
                    Family("f1",Edit(FRED_0="%FRED_1%"),
                          Task("f1_t1", Edit(TASK_VAR="3",FRED="%FRED_0%")),
                          Task("f1_t2", Edit(TASK_VAR="3",FRED="%FRED_0%"))
                    ),
                    Family("f2",Edit(FRED_0="%FRED_1%"),
                          Task("f2_t1", Edit(TASK_VAR="3",FRED="%FRED_0%")),
                          Task("f2_t2", Edit(TASK_VAR="3",FRED="%FRED_0%"))
                    )
                )
            )
    return defs;

def test_ecflow_1213(defs):
    s1 = defs.find_node_path("suite","s1")
    assert s1 == "/s1"," expected to find suite s1"

    f1 = defs.find_node_path("family","f1")
    assert f1 == "/s1/f1"," expected to find family f1"
  
    f2 = defs.find_node_path("family","f2")
    assert f2 == "/s1/f2"," expected to find family f2"

    f1_t1 = defs.find_node_path("task","f1_t1")
    f1_t2 = defs.find_node_path("task","f1_t2")
    assert f1_t1 == "/s1/f1/f1_t1"," expected to find task f1_t1"
    assert f1_t2 == "/s1/f1/f1_t2"," expected to find task f1_t2"
    
    f2_t1 = defs.find_node_path("task","f2_t1") 
    f2_t2 = defs.find_node_path("task","f2_t2")
    assert f2_t1 == "/s1/f2/f2_t1"," expected to find task f2_t1"
    assert f2_t2 == "/s1/f2/f2_t2"," expected to find task f2_t2"
    
    # find_node
    s1 = defs.find_node("suite","/s1")
    assert s1 is not None and isinstance(s1, Suite), "expected find find suite s1"
    
    f1 = defs.find_node("family","/s1/f1")
    assert f1 is not None and isinstance(f1, Family)," expected to find family f1"
  
    f2 = defs.find_node("family","/s1/f2")
    assert f2 is not None and isinstance(f2, Family)," expected to find family f2"

    f1_t1 = defs.find_node("task","/s1/f1/f1_t1")
    f1_t2 = defs.find_node("task","/s1/f1/f1_t2")
    assert f1_t1 is not None and isinstance(f1_t1, Task)," expected to find task f1_t1"
    assert f1_t2 is not None and isinstance(f1_t2, Task)," expected to find task f1_t2"
    
    f2_t1 = defs.find_node("task","/s1/f2/f2_t1")
    f2_t2 = defs.find_node("task","/s1/f2/f2_t2")
    assert f2_t1 is not None and isinstance(f2_t1, Task)," expected to find task f2_t1"
    assert f2_t2 is not None and isinstance(f2_t1, Task)," expected to find task f2_t2"


if __name__ == "__main__":   
    
    Test.print_test_start(os.path.basename(__file__))
 
    defs = create_defs()
    
    test_ecflow_1213(defs)
    
    s1 = defs.find_suite("s1")
    assert s1 is not None, "expected find find suite s1"
    assert "s1" in defs, "expected find find suite s1"

    sx = defs.find_suite("sx")
    assert sx is None, "expected not to find suite sx"
    
    f1 = s1.find_family("f1")
    assert f1 is not None, "expected find find family f1"
    f1 = s1.find_node("f1")
    assert f1 is not None, "expected find find family f1"

    fx = s1.find_family("fx")
    assert fx is None, "expected not to find family fx"
    fx = s1.find_node("fx")
    assert fx is None, "expected not to find family fx"

    f2 = s1.find_family("f2")
    assert f2 is not None, "expected find find family f2"
    f2 = s1.find_node("f2")
    assert f2 is not None, "expected find find family f2"
    
    f2_t1 = f2.find_task("f2_t1")
    assert f2_t1 is not None, "Expected to find task"

    f2_t2 = f2.find_task("f2_t2")
    assert f2_t2 is not None, "Expected to find task"
    f2_t2 = f2.find_node("f2_t2")
    assert f2_t2 is not None, "Expected to find task"

    f2_tx = f2.find_task("f2_tx")
    assert f2_tx is None, "Expected not to find task"
    f2_tx = f2.find_node("f2_tx")
    assert f2_tx is None, "Expected not to find task"

    tasks = defs.get_all_tasks()
    assert len(tasks) == 4, "Expected four tasks, but found " + str(len(tasks))
    for task in tasks:
        # test find variable
        var = task.find_parent_variable("TASK_VAR")
        assert not var.empty()
        assert var.value() == "3"
        
        var = task.find_parent_variable_sub_value("FRED")
        assert var == "XX"

        var = task.find_parent_variable_sub_value("ZZZZ")
        assert var == ""

        var = task.find_variable("TASK_VAR")
        assert not var.empty()
        assert var.value() == "3"
        
        var = task.find_parent_variable("SUITE_VAR");
        assert not var.empty()
        assert var.value() == "1"
         
        var = task.find_parent_variable("DEFS_VAR")
        assert not var.empty()
        assert var.value() == "0"

        var = task.find_parent_variable("MADE_UP_VAR")
        assert var.empty()

        if task.name() == "f1_t1":
            node = task.find_node_up_the_tree("f1")
            assert node is not None
            assert node.get_abs_node_path() == "/s1/f1"
            
            node = task.find_node_up_the_tree("s1")
            assert node is not None
            assert node.get_abs_node_path() == "/s1"

            node = task.find_node_up_the_tree("freddd")
            assert node is None
            
    print("All Tests pass")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
from ecflow import Alias, AttrType, Autocancel, CheckPt, ChildCmdType, Client, Clock, Cron, DState, Date, Day, Days, \
                  Defs, Ecf, Event, Expression, Family, FamilyVec, File, Flag, FlagType, FlagTypeVec, InLimit, \
                  JobCreationCtrl, Label, Late, Limit, Meter, Node, NodeContainer, NodeVec, PartExpression, PrintStyle, \
                  Repeat, RepeatDate, RepeatDateTime, RepeatDay, RepeatEnumerated, RepeatInteger, RepeatString, SState, State, Style, \
                  Submittable, Suite, SuiteVec, Task, TaskVec, Time, TimeSeries, TimeSlot, Today, UrlCmd, Variable, \
                  VariableList, Verify, WhyCmd, ZombieAttr, ZombieType, ZombieUserActionType, Trigger, Complete, Edit, Defstatus
import os 
import unittest 

class TestAddSuiteFamilyTask0(unittest.TestCase):
    def setUp(self):
        defs = Defs()
        s = Suite('s1')
        f = Family('f1')
        t = Task('t1')
        defs.add_suite(s)
        s.add_family(f)
        f.add_task(t)
        self.defs = defs

    def test_0(self):
        defs = Defs()
        defs.add_suite('s1').add_family('f1').add_task('t1')
        self.assertEqual(self.defs, defs, "defs not equal")
        
    def test_1(self):
        defs = Defs().add(
            Suite('s1').add(
                Family('f1').add(
                    Task('t1'))))
        self.assertEqual(self.defs, defs, "defs not equal")

    def test_2(self):
        defs = Defs(
            Suite('s1',
                  Family('f1',
                         Task('t1'))))
        self.assertEqual(self.defs, defs, "defs not equal")

    def test_3(self):
        defs = Defs() + (Suite('s1') + (Family('f1') + Task('t1')))
        self.assertEqual(self.defs, defs, "defs not equal")

    def test_3(self):
        defs = Defs() + (Suite('s1') + (Family('f1') + Task('t1')))
        self.assertEqual(self.defs, defs, "defs not equal")
          
class TestAddSuiteFamilyTask(unittest.TestCase):
    def setUp(self):
        defs = Defs()            # create an empty definition
        suite = defs.add_suite("s1")    # create a suite and add it to the defs
        family = suite.add_family("f1") # create a family and add it to suite
        for i in [ "a", "b", "c" ]:     # create task ta,tb,tc
            family.add_task( "t" + i)   # create a task and add to family
        defs.save_as_defs("test.def")   # save defs to file "test.def"  
         
        self.defs = defs
         
    def test_me(self):
        with Defs() as defs:
            with defs.add_suite("s1") as suite:
                with suite.add_family("f1") as family:
                    for i in [ "a", "b", "c" ]:      
                        family.add_task( "t" + i)    
        defs.save_as_defs("test.def")    
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)
        self.assertEqual(self.defs, defs, "defs not equal")
         
    def test_me2(self):
        defs = Defs().add( 
            Suite("s1").add(
                Family("f1").add(
                    [ Task("t{0}".format(t)) 
                      for t in ("a", "b", "c")] )))
        defs.save_as_defs("test.def")   # save defs to file "test.def"  
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)
        self.assertEqual(self.defs, defs, "defs not equal")
 
    def test_me3(self):
        defs = Defs()
        defs += Suite("s1") 
        defs.s1 += Family("f1") 
        defs.s1.f1 += [ Task("t{0}".format(t)) 
                        for t in ("a", "b", "c")] 
        defs.save_as_defs("test.def")   # save defs to file "test.def"  
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)
        self.assertEqual(self.defs, defs, "defs not equal")
        
    def test_me4(self):
        defs = Defs(
                Suite("s1",
                    Family("f1",
                        [ Task("t{0}".format(t)) for t in ("a", "b", "c")])))
        defs.save_as_defs("test.def")   # save defs to file "test.def"  
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)
        self.assertEqual(self.defs, defs, "defs not equal")
 
    def test_me5(self):
        defs = Defs().add(   
          [ Suite("s{0}".format(i)).add( 
              [ Family("f{0}".format(i)).add( 
                  [ Task("t{0}".format(i)) for i in range(1,6)] ) 
                for i in range(1,7)]  )
            for i in range(1,6) ] )
        assert len(defs)==5, " expected 6 suites but found " + str(len(defs))
        for suites in defs:
            assert len(suites)==6, " expected 6 families but found " + str(len(suites))
            for fam in suites:
                assert len(fam)==5, " expected 6 tasks but found " + str(len(fam))
                
    def test_me6(self):
        defs = Defs(
                [ Suite("s{0}".format(i),
                    [ Family("f{0}".format(i),
                        [ Task("t{0}".format(i)) for i in range(1,6)] )
                    for i in range(1,7)]  )
                for i in range(1,6) ] )
        assert len(defs)==5, " expected 5 suites but found " + str(len(defs))
        for suites in defs:
            assert len(suites)==6, " expected 6 families but found " + str(len(suites))
            for fam in suites:
                assert len(fam)==5, " expected 5 tasks but found " + str(len(fam))
             
    def tearDown(self):
        unittest.TestCase.tearDown(self)
        try: os.remove("test.def")
        except: pass
         
class TestAddMeterEventLabel(unittest.TestCase):
    def setUp(self):
        defs = Defs()
        suite = Suite("s1")
        task = Task("t1")
        defs.add_suite(suite)
        suite.add_task(task)
        task.add_event( 2 )                      # event reference with 2
        task.add_event("wow")                    # event reference with name "wow"
        task.add_event( 10,"Eventname2" )        # event referenced with name "Eventname2"
        task.add_meter( "metername3",0,100 )     # name, min, max
        task.add_label( "label_name4", "value" ) # name, value
         
        self.defs = defs
         
    def test_alternative(self):
        defs = Defs().add(
            Suite("s1").add(
                Task("t1").add(
                    Event(2),
                    Event("wow"),
                    Event(10,"Eventname2" ),
                    Meter("metername3",0,100),
                    Label("label_name4", "value"))))
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs,self.defs, "expected defs to be the same")
 
    def test_alternative1(self):
        defs = Defs() + Suite("s1")
        defs.s1 += Task("t1") + Event(2) + Event("wow") + \
                    Event(10,"Eventname2" ) + Meter("metername3",0,100) + \
                    Label("label_name4", "value") 
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
         
    def test_alternative2(self):
        with Defs() as defs:
            with defs.add_suite("s1") as suite:
                with suite.add_task("t1") as t1:
                    t1 += [ Event(2),
                           Event("wow"),
                           Event(10,"Eventname2" ),
                           Meter("metername3",0,100),
                           Label("label_name4", "value") ]
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative3(self):
        defs = Defs( 
                Suite('s1',
                    Task("t1",
                        Event(2),
                        Event("wow"),
                        Event(10,"Eventname2" ),
                        Meter("metername3",0,100),
                        Label("label_name4", "value")))) 
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
            
class TestAddLimitInlimit(unittest.TestCase):
     
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        s1.add_limit( "limitName4", 10 ) # name, maximum token
        f1 = s1.add_family("f1")
        f1.add_inlimit( "limitName4","/s1/f1",2) # limit name, path to limit, tokens consumed
        for i in range(1,4):
            f1.add_task( "t{0}".format(i))
             
        self.defs = defs
 
    def test_alternative(self):
        defs = Defs().add(
            Suite("s1").add(
                Limit("limitName4", 10),
                Family("f1").add(
                    InLimit("limitName4","/s1/f1",2),
                    [ Task("t{0}".format(t)) for t in range(1,4) ]
                    )))
 
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")        
 
    def test_alternative1(self):
        defs = Defs() + Suite("s1") 
        defs.s1 += [ Limit("limitName4", 10),Family("f1") ]
        defs.s1.f1 += [ InLimit("limitName4","/s1/f1",2),
                        [ Task("t{0}".format(t)) for t in range(1,4) ] ]
 
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")        
         
    def test_alternative2(self):
        with Defs() as defs:
            with defs.add_suite("s1") as s1:
                s1.add_limit( "limitName4", 10 ) # name, maximum token
                with s1.add_family("f1") as f1:
                    f1.add_inlimit( "limitName4","/s1/f1",2) # limit name, path to limit, tokens consumed
                    f1 += [ Task("t{0}".format(t)) for t in range(1,4) ]
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative3(self):
        defs = Defs( 
            Suite("s1",
                Limit("limitName4", 10),# name, maximum token
                Family("f1",
                    InLimit("limitName4","/s1/f1",2), # limit name, path to limit, tokens consumed
                    [ Task("t{0}".format(t)) for t in range(1,4) ] )))
                     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
 
 
class TestAddVariable(unittest.TestCase):
     
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        s1.add_variable("HELLO","world") # name, value
        s1.add_variable({ "NAME":"value", "NAME2":"value2", "NAME3":"value3", "NAME4":4 }  )
        s1.add_variable(Variable("FRED","bloggs"))
        s1.add_variable(Variable("BILL","1"))
         
        defs.s1.sort_attributes("variable"); # sort since with dictionary order of addition is arbitary
        self.defs = defs

    def test_alternative0(self):
         
        defs = Defs(Suite("s1",HELLO="world",FRED="bloggs",BILL=1,NAME="value",NAME2="value2"))
        defs.s1.add_variable('NAME4',4)
        defs.s1 += Edit(NAME3="value3")
 
        defs.s1.sort_attributes("variable");
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
 
    def test_alternative(self):
         
        defs = Defs().add(
                      Suite("s1").add(
                          Edit(HELLO="world",FRED="bloggs",BILL=1,NAME="value",NAME2="value2",NAME3="value3",NAME4=4 )
                          ))
 
        defs.s1.sort_attributes("variable");
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
 
    def test_alternative1(self):
         
        defs = Defs() + Suite("s1",Edit(HELLO="world"))
        defs.s1 += [ Edit({ "NAME":"value", "NAME2":"value2", "NAME3":"value3", "NAME4":4 }, BILL=1),
                     Edit(FRED="bloggs") ]
 
        defs.s1.sort_attributes("variable");
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative2(self):
         
        defs = Defs(
                Suite("s1",
                    { "HELLO":"world", "NAME":"value", "NAME2":"value2", 
                     "NAME3":"value3", "NAME4":4, "BILL":1, "FRED":"bloggs" }))
 
        defs.s1.sort_attributes("variable");
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

class TestAddTrigger(unittest.TestCase):
     
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        t1 = s1.add_task("t1")
        t2 = s1.add_task("t2")
        t2.add_trigger( "t1 == active and t3 == aborted" )
        t2.add_complete( "t3 == complete" )
        t3 = s1.add_task("t3")
        self.defs = defs

    def test_alternative0(self):
        
        defs = Defs(
                Suite("s1",
                    Task("t1"),
                    Task("t2",
                        Trigger("t1 == active and t3 == aborted"),
                        Complete("t3 == complete")),
                    Task("t3")))
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative(self):
        
        defs = Defs().add(
                 Suite("s1").add(
                    Task("t1"),Task("t2").add(
                        Trigger("t1 == active and t3 == aborted"),
                        Complete("t3 == complete")),
                    Task("t3")))
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
 
    def test_alternative1(self):
         
        defs = Defs() + Suite("s1")
        defs.s1 += [ Task("t{0}".format(i)) for i in range(1,4) ]
        defs.s1.t2 += [ Trigger("t1 == active and t3 == aborted"),
                        Complete("t3 == complete") ]
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

class TestAddTaskChain(unittest.TestCase):

    def setUp(self):
        defs = Defs(
                Suite("s1",
                    Task("t1"),
                    Task("t2",Trigger( "t1 == complete" )),
                    Task("t3",Trigger( "t2 == complete" )),
                    Task("t4",Trigger( "t3 == complete" ))))
        self.defs = defs
        
    def test_alternative(self):
        defs = Defs() + Suite("s1")
        defs.s1 += [ Task("t1"),Task("t2"),
                     Task("t3"),Task("t4") ]
        defs.s1.t2 += Trigger( ["t1"] )
        defs.s1.t3 += Trigger( ["t2"] )
        defs.s1.t4 += Trigger( ["t3"] )

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

    def test_alternative1(self):
        defs = Defs() + Suite("s1")
        defs.s1 += [ Task("t1"),Task("t2"),Task("t3"),Task("t4") ]
        defs.s1.t2 += Trigger( [ defs.s1.t1 ] )
        defs.s1.t3 += Trigger( [ defs.s1.t2 ] )
        defs.s1.t4 += Trigger( [ defs.s1.t3 ] )

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative1(self):
        defs = Defs(Suite("s1"))
        defs.s1 >> Task("t1") >> Task("t2") >> Task("t3") >> Task("t4")

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
class TestAddReverseTaskChain(unittest.TestCase):

    def setUp(self):
        defs = Defs(
                Suite("s1",
                    Task("t1", Trigger( "t2 == complete" )),
                    Task("t2", Trigger( "t3 == complete" )),
                    Task("t3", Trigger( "t4 == complete" )),
                    Task("t4")))
        self.defs = defs
        
    def test_alternative1(self):
        defs = Defs() + Suite("s1")
        defs.s1 << Task("t1") << Task("t2") << Task("t3") << Task("t4")

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
class TestAddLargeTrigger(unittest.TestCase):
    
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        t1 = s1.add_task("t1")
        t2 = s1.add_task("t2")
        t3 = s1.add_task("t3")
        t3.add_part_trigger( "t1 == complete")
        t3.add_part_trigger( "t2 == active", True) # here True means add as 'AND'
        t3.add_part_trigger( "t2 == aborted", False) # here False means add as 'OR'
        self.defs = defs

    def test_alternative0(self):
        defs = Defs( 
                Suite("s1",
                    Task("t1"),
                    Task("t2"),
                    Task("t3",
                        Trigger("t1 == complete"), 
                        Trigger("t2 == active"),
                        Trigger("t2 == aborted",False))))
        
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative(self):
        defs = Defs().add(
                Suite("s1").add(
                    Task("t1"),
                    Task("t2"),
                    Task("t3").add(
                        Trigger("t1 == complete"), 
                        Trigger("t2 == active"),
                        Trigger("t2 == aborted",False),
                        )
                    ))
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

    def test_alternative1(self):
         
        defs = Defs() + Suite("s1")
        defs.s1 += [ Task("t{0}".format(i)) for i in range(1,4) ]
        defs.s1.t3 += [ Trigger("t1 == complete"),
                        Trigger("t2 == active"),
                        Trigger("t2 == aborted",False) ]
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
class TestAddTimeDependencies(unittest.TestCase):
    
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        t1 = s1.add_task("date")
        t1.add_date(1, 0, 0)       # first of every month and every year
        t1.add_date(Date("2.*.*")) # second of every month and every year
        t1.add_date(28,2,2026)     # 28 february 2026

        t2 = s1.add_task("day")
        t2.add_day("monday")
        t2.add_day(Days.tuesday)
        
        t3 = s1.add_task("time")
        t3.add_time("+00:30")                     # 30 minutes after suite has begun
        t3.add_time("+00:30 20:00 01:00")         # 00:30,01:30,02:30....07:30 after suite start
              
        t3.add_time(0, 59, True)                  #  00:59 - 59 minutes past midnight
        t3.add_time(Time(TimeSlot(20, 10)))       #  20:10 - 10 minutes pas eight
        t3.add_time(Time(TimeSlot(20, 20), True)) # +20:20 - 20 minutes and 20 hours, after suite start

        start = TimeSlot(0, 0)
        finish = TimeSlot(23, 0)
        incr = TimeSlot(0, 30)
        time_series = TimeSeries(start, finish, incr, True)        
        t3.add_time(Time(time_series))
        t3.add_time(Time(0, 10))                  #  00:10 
        t3.add_time(Time("+00:40"))               # +00:40
        t3.add_time(Time("+00:40 20:00 01:00"))   # 00:40,01:40,02:30...17:40 after suite start
        
        cron = Cron()
        cron.set_week_days( [0,1,2,3,4,5,6] )
        cron.set_days_of_month( [1,2,3,4,5,6] )
        cron.set_months( [1,2,3,4,5,6] )
        cron.set_time_series( "+00:00 23:00 00:30" )
        s1.add_task("cron").add_cron( cron );
        
        self.defs = defs
        
    def test_alternative0(self):
        start = TimeSlot(0, 0)
        finish = TimeSlot(23, 0)
        incr = TimeSlot(0, 30)
        time_series = TimeSeries(start, finish, incr, True)
        
        defs = Defs( 
                Suite("s1",
                    Task("date",
                        Date(1, 0, 0),                   # first of every month and every year
                        Date("2.*.*"),                   # second of every month and every yea
                        Date(28,2,2026)),                # 28 february 2026
                    Task("day",
                        Day("monday"),
                        Day(Days.tuesday)),
                    Task("time",
                        Time("+00:30"),                  # 30 minutes after suite has begun
                        Time("+00:30 20:00 01:00"),      # 00:30,01:30,02:30....07:30 after suite start
                        Time(0, 59, True),               # 00:59 - 59 minutes past midnight
                        Time(TimeSlot(20, 10)),          # 20:10 - 10 minutes pas eight
                        Time(TimeSlot(20, 20), True),    # +20:20 - 20 minutes and 20 hours, after suite start
                        Time(time_series),
                        Time(0, 10),
                        Time("+00:40"),
                        Time("+00:40 20:00 01:00")),
                    Task("cron",
                        Cron("+00:00 23:00 00:30",days_of_week=[0,1,2,3,4,5,6],days_of_month=[1,2,3,4,5,6],months=[1,2,3,4,5,6]))))

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative(self):
        start = TimeSlot(0, 0)
        finish = TimeSlot(23, 0)
        incr = TimeSlot(0, 30)
        time_series = TimeSeries(start, finish, incr, True)
        
        cron = Cron()
        cron.set_week_days( [0,1,2,3,4,5,6] )
        cron.set_days_of_month( [1,2,3,4,5,6] )
        cron.set_months( [1,2,3,4,5,6] )
        cron.set_time_series( "+00:00 23:00 00:30" )

        defs = Defs().add(
            Suite("s1").add(
                Task("date").add(
                    Date(1, 0, 0),
                    Date("2.*.*"),
                    Date(28,2,2026)),
                Task("day").add(
                    Day("monday"),
                    Day(Days.tuesday)),
                Task("time").add(
                    Time("+00:30"),
                    Time("+00:30 20:00 01:00"),
                    Time(0, 59, True),
                    Time(TimeSlot(20, 10)),
                    Time(TimeSlot(20, 20), True),
                    Time(time_series),
                    Time(0, 10),
                    Time("+00:40"),
                    Time("+00:40 20:00 01:00")),
                Task("cron").add(
                    cron)))

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative1(self):
        start = TimeSlot(0, 0)
        finish = TimeSlot(23, 0)
        incr = TimeSlot(0, 30)
        time_series = TimeSeries(start, finish, incr, True)
        
        defs = Defs() + ( Suite("s1") + Task("date") + Task("day") + Task("time") + Task("cron"))
        defs.s1.date += [ Date(1, 0, 0), Date("2.*.*"), Date(28,2,2026) ]
        defs.s1.day += [ Day("monday"), Day(Days.tuesday) ]
        defs.s1.time += [ Time("+00:30"), Time("+00:30 20:00 01:00"), Time(0, 59, True),
                          Time(TimeSlot(20, 10)), Time(TimeSlot(20, 20), True),
                          Time(time_series), Time(0, 10), Time("+00:40"),
                          Time("+00:40 20:00 01:00") ]
        defs.s1.cron += Cron("+00:00 23:00 00:30",
                             days_of_week=[0,1,2,3,4,5,6],
                             days_of_month=[1,2,3,4,5,6],
                             months=[1,2,3,4,5,6])  

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
class TestAddDefStatus(unittest.TestCase):
    
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        s1.add_task("t1").add_defstatus(Defstatus("complete"))
        s1.add_task("t2").add_defstatus(DState.complete)
        
        self.defs = defs

    def test_alternative0(self):
        defs = Defs( 
                Suite("s1",
                    Task("t1",Defstatus("complete")),
                    Task("t2",Defstatus(DState.complete))))
        
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

    def test_alternative(self):
        defs = Defs().add(
            Suite("s1").add(
                Task("t1").add(Defstatus("complete")),
                Task("t2").add(Defstatus(DState.complete))))
        
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")

    def test_alternative1(self):
        defs = Defs() + ( Suite("s1") +  Task("t1") + Task("t2") )
        defs.s1.t1 += Defstatus("complete") 
        defs.s1.t2 += Defstatus(DState.complete) 
        
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
   
class TestAddAutocancel(unittest.TestCase):
    
    def setUp(self):
        defs = Defs()
        s1 = defs.add_suite("s1") 
        s1.add_task("t1").add_autocancel(3)                                # delete task after 3 days after completion 
        s1.add_task("t2").add_autocancel(1, 10, True)                      # delete task 1hr 10 min after task completion
        s1.add_task("t3").add_autocancel(TimeSlot(2,10), True)             # delete task 2hr 10 min after task completion 
        s1.add_task("t4").add_autocancel(Autocancel(1))                    # delete task after 1 day after task completion
        s1.add_task("t5").add_autocancel(Autocancel( 18, 10, False))       # delete task at 6:10pm once it has completed
        s1.add_task("t6").add_autocancel(Autocancel(TimeSlot(2,10), False))# delete task at 2:10am once it has completed

        self.defs = defs

    def test_alternative0(self):
        defs = Defs( 
                Suite("s1",
                    Task("t1", Autocancel(3)),                    # delete task after 3 days after completion 
                    Task("t2", Autocancel(1, 10, True)),          # delete task 1hr 10 min after task completion
                    Task("t3", Autocancel(TimeSlot(2,10), True)), # delete task 2hr 10 min after task completion
                    Task("t4", Autocancel(1)),                    # delete task after 1 day after task completion
                    Task("t5", Autocancel(18, 10, False)),        # delete task at 6:10pm once it has completed
                    Task("t6", Autocancel(2, 10, False))))        # delete task at 2:10am once it has completed

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative(self):
        defs = Defs().add(
            Suite("s1").add(
                Task("t1").add(Autocancel(3)),
                Task("t2").add(Autocancel(1, 10, True)),
                Task("t3").add(Autocancel(TimeSlot(2,10), True)),
                Task("t4").add(Autocancel(1)),
                Task("t5").add(Autocancel(18, 10, False)),
                Task("t6").add(Autocancel(2, 10, False))))

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")


    def test_alternative1(self):
        defs = Defs() + ( Suite("s1") + [ Task("t{0}".format(i)) for i in range(1,7)] )
        defs.s1.t1 += Autocancel(3) 
        defs.s1.t2 += Autocancel(1, 10, True) 
        defs.s1.t3 += Autocancel(TimeSlot(2,10), True) 
        defs.s1.t4 += Autocancel(1)     
        defs.s1.t5 += Autocancel(18, 10, False) 
        defs.s1.t6 += Autocancel(2, 10, False)       
         
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
   
class TestAddRepeat(unittest.TestCase):
    def setUp(self):
        def add_tasks(fam):
            for i in range(1,3):
                fam.add_task(Task("t{0}".format(i)))
                
        defs = Defs()
        s1 = defs.add_suite("s1")
        f1 = s1.add_family("f1")
        f1.add_repeat( RepeatDate("YMD",20100111,20100115,2) )
        add_tasks(f1)

        f2 = s1.add_family("f2")
        f2.add_repeat( RepeatInteger("count",0,100,2)  )
        add_tasks(f2)

        f3 = s1.add_family("f3")
        f3.add_repeat( RepeatEnumerated("enum",["red", "green", "blue" ] )  )
        add_tasks(f3)

        f4 = s1.add_family("f4")
        f4.add_repeat( RepeatString("enum",["a", "b", "c" ] )  )
        add_tasks(f4)

        f5 = s1.add_family("f5")
        f5.add_repeat( RepeatDay(1)  )
        add_tasks(f5)

        f6 = s1.add_family("f6")
        f6.add_repeat(RepeatDateTime("DT", "20100111T000000", "20100115T000000", "48:00:00"))
        add_tasks(f6)
          
        self.defs = defs
     

    def test_alternative0(self):
        defs = Defs( 
                Suite("s1",
                    Family("f1",
                       RepeatDate("YMD",20100111,20100115,2),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f2",
                       RepeatInteger("count",0,100,2),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f3",
                       RepeatEnumerated("enum",["red", "green", "blue" ] ),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f4",
                        RepeatString("enum",["a", "b", "c" ] ),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f5",
                        RepeatDay(1),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f6",
                        RepeatDateTime("DT", "20100111T000000", "20100115T000000", "48:00:00"),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] )))

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
    def test_alternative1(self):
        defs = Defs().add(
                Suite("s1").add(
                    Family("f1").add(
                       RepeatDate("YMD",20100111,20100115,2),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f2").add(
                       RepeatInteger("count",0,100,2),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f3").add(
                       RepeatEnumerated("enum",["red", "green", "blue" ] ),
                       [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f4").add(
                        RepeatString("enum",["a", "b", "c" ] ),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f5").add(
                        RepeatDay(1),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] ),
                    Family("f6").add(
                        RepeatDateTime("DT", "20100111T000000", "20100115T000000", "48:00:00"),
                        [ Task("t{0}".format(i)) for i in range(1,3) ] )))

        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
 
    def test_alternative2(self):
        defs = Defs() + Suite("s1") 
        defs.s1 += [ Family("f{0}".format(i)).add(
                      [ Task("t{0}".format(i)) for i in range(1,3) ]) 
                    for i in range(1,7) ]
        defs.s1.f1 += RepeatDate("YMD",20100111,20100115,2)
        defs.s1.f2 += RepeatInteger("count",0,100,2) 
        defs.s1.f3 += RepeatEnumerated("enum",["red", "green", "blue" ] ) 
        defs.s1.f4 += RepeatString("enum",["a", "b", "c" ] ) 
        defs.s1.f5 += RepeatDay(1)
        defs.s1.f6 += RepeatDateTime("DT", "20100111T000000", "20100115T000000", "48:00:00")
  
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")
        
        
class TestAddLate(unittest.TestCase):
    def setUp(self):
        defs = Defs()
        suite = defs.add_suite('s1')
        
        late = Late()
        late.submitted( 20,10 )         # hour, min
        late.active( 2, 10 )            # hour, min
        late.complete( 3, 10, True)     # hour, min, relative
        suite.add_task('t1').add_late(late)
        self.defs = defs
 
    def test_1(self):
        
        # Can also pass late into the Task constructor
        defs = Defs(
                Suite('s1',
                    Task('t1',
                        Late(submitted='20:10',active='02:10',complete='+03:10'))))
     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")   
        
    def test_2(self):
        
        # Can also pass late into the Task constructor
        defs = Defs() + (Suite('s1') + Task('t1'))
        defs.s1.t1 += Late(submitted='20:10',active='02:10',complete='+03:10')
     
        Ecf.set_debug_equality(True)
        equals = (self.defs == defs)
        Ecf.set_debug_equality(False)      
        self.assertEqual(defs, self.defs, "expected defs to be the same")   
        
        
class Deadlock(unittest.TestCase):
    def setUp(self):
        defs = Defs().add(
        Suite("dead_lock").add(
            Family("family").add(
               Task("t1").add( Trigger("t2 == complete")),
               Task("t2").add( Trigger("t1 == complete")))))
        
        self.defs = defs
        
    def test_me(self):
        defs = Defs(
            Suite("dead_lock",
                Family('family',
                    Task('t1', 
                        Trigger("t2 == complete")),
                    Task('t2', 
                        Trigger("t1 == complete")))))
        
        self.assertEqual(defs, self.defs, "expected defs to be the same")
       
        
if __name__ == "__main__":
    unittest.main()
    print("All Tests pass")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import unittest

class TestConfluence(unittest.TestCase):
    def setUp(self):
        conflunce_base_url = "https://software-test.ecmwf.int/wiki"
#         conflunce_base_url = "https://confluence.ecmwf.int"
        user = "deploy"  
        password = "deploy2013"  
#         user = "ma0"
#         password = "chacchok1"
        
        import sys
        sys.path.append("/var/tmp/ma0/workspace/admin/2.0")
        import rest
        self.c = rest.Confluence(conflunce_base_url,user,password)

    def test_add_attachment_id(self):
        title = "Releases"
        space_key_list = [ "EMOS", "ECFLOW", "MAGP", "METV", "ECCUI", "ECC" ]
        for space_key in space_key_list:
            page_id = self.c.get_page_id(space_key,title) 
            if page_id is None:
                print("Could not find page id for space_key ",space_key," and title:",title)
            else:
                print("space key:",space_key," page id:",page_id)

        comment = "production release"
        file = "/var/tmp/ma0/workspace/ecflow/libs/pyext/samples/test.tar.gz"
        for space_key in space_key_list:
            page_id = self.c.get_page_id(space_key,title) 
            
            attachment = self.c.get_attachment_id(page_id,file)
            if attachment is not None:
                print("space key:",space_key," allready has an attachment for file ",file)
            else:
                self.c.create_attachment(page_id,comment,file)

        
if __name__ == '__main__':
    unittest.main()
    print("All Tests pass")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

#  Test the error and waning checks defined in the defs work
import os
from ecflow import  Defs, Limit, InLimit, Client, debug_build
import ecflow_test_util as Test

if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
     
    defs = Defs()
    suite = defs.add_suite("s1");
    suite.add_task("t1").add_trigger("t2 == active)")   
    theCheckValue = defs.check();
    print("Message: '" + theCheckValue + "'")
    assert len(theCheckValue) != 0,  "Expected Error: triggers fail parse, miss-matched brackets in expression."
 
    # The number of tokens specified on the in-limit must be less than or equal to those specified on the LIMIT
    defs = Defs()
    suite = defs.add_suite("s1");
    suite.add_limit( Limit("disk", 50) );
    family = suite.add_family("anon");
    family.add_inlimit( InLimit("disk","/s1",100) )
    family.add_task( "t1" );
    theCheckValue = defs.check();
    print("Message: '" + theCheckValue + "'")
    assert len(theCheckValue) != 0,  "Expected Error: since inlimit value('100') is greater than the LIMIT value('50')"  


    # When a path is specified on the in-limit we search for the limit on that path, otherwise the extern's
    defs = Defs()
    suite = defs.add_suite("s1");
    family = suite.add_family("anon");
    family.add_inlimit( InLimit("disk","/s1",100) )
    family.add_task( "t1" );
    theCheckValue = defs.check();
    print("Message: '" + theCheckValue + "'")
    assert len(theCheckValue) != 0,  "Expected warning: since inlimit PATH(/s1:disk) reference a limit('disk') that does not exist"  
    
    # Add as extern and repeat the check
    defs.add_extern("/s1:disk")
    assert len(defs.check()) == 0,  "Expected no warnings, since extern specified: "  + defs.check()


    # When no path is specified on the in-limit, we search for the limit up the node tree, otherwise the extern's
    defs = Defs()
    suite = defs.add_suite("s1");
    family = suite.add_family("anon");
    family.add_inlimit( InLimit("disk","",100) )
    family.add_task( "t1" );
    theCheckValue = defs.check();
    print("Message: '" + theCheckValue + "'")
    assert len(theCheckValue) != 0,  "Expected warning: since inlimit SHOULD reference a limit('disk') somewhere UP parent hierarchy"  
    
    # Add as extern and repeat the check
    defs.add_extern("disk")
    assert len(defs.check()) == 0,  "Expected no warnings, since extern specified"  
    
    print("All Tests pass")
    


#!/usr/bin/env python2.7
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
import ecflow
import argparse # for argument parsing  
import sys   

if __name__ == "__main__":
    
    DESC = """Will list variables for any node
              Usage:
                Example1: List all the server variables
                   ListVariables.py --host cca --port 4141 --path /
                   
                Example2: List the variables for the given node
                   ListVariables.py --host cca --port 4141 --path /path/to/node
            """    
    PARSER = argparse.ArgumentParser(description=DESC,  
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    PARSER.add_argument('--host', default="localhost",   
                        help="The name of the host machine, defaults to 'localhost'")
    PARSER.add_argument('--port', default="3141",   
                        help="The port on the host, defaults to 3141")
    PARSER.add_argument('--path', default="/",   
                        help="The path to the node. i.e /suite/family/task")
    ARGS = PARSER.parse_args()
    #print ARGS    
    
    # ===========================================================================
    CL = ecflow.Client(ARGS.host, ARGS.port)
    try:
        CL.ping() 

        # get the incremental changes, and merge with defs stored on the Client 
        CL.sync_local()
        
        # check to see if definition exists in the server
        defs = CL.get_defs()
        if len(defs) == 0 :
            print("No suites found, exiting...")
            sys.exit(0) 
            
        # print defs;
        if ARGS.path == "/":
            for var in defs.server_variables:  print("edit " + var.name() + " '" + var.value() + "'")
            for var in defs.user_variables:    print("edit " + var.name() + " '" + var.value() + "'")
        else:
            node = defs.find_abs_node(ARGS.path)
            if node is None:  
                print("No node found at path " + ARGS.path)
            else:
                for var in node.variables:     print("edit " + var.name() + " '" + var.value() + "'")

    except RuntimeError as ex:
        print("Error: " + str(ex))
        print("Check host and port number are correct.")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import os
from ecflow import Defs, Suite, Variable, Limit, InLimit, Task, PartExpression, \
                   Event, Meter, Label, Queue, RepeatInteger, RepeatEnumerated, RepeatDate, RepeatDateTime, \
                   RepeatDateList, RepeatString, TimeSlot, TimeSeries, Today, Time, Date, Day, Days, Cron, Autocancel, \
                   Late, DState, Clock, ChildCmdType, ZombieType, ZombieAttr, ZombieUserActionType, Client, debug_build
import ecflow_test_util as Test

if __name__ == "__main__":

    Test.print_test_start(os.path.basename(__file__))
    
    #
    # Add Nodes functional way
    #
    defs = Defs()
    defs.add_suite("s1").add_task("t1").add_variable("var","v")
    defs.add_suite("s2").add_family("f1").add_task("t1").add_variable("var","v")
    defs.add_suite("s3").add_family("f1").add_family("f2").add_task("t1").add_variable("var","v")
    assert len(defs) == 3,"Expected 3 suites"    


    # add and delete variables
    a_dict = { "name":"value", "name2":"value2", "name3":"value3", "name4":"value4" }
    suite = Suite("s1")
    suite.add_variable(Variable("ECF_HOME","'/tmp/'")) \
         .add_variable("Fred", 1)\
         .add_variable("ECF_URL_BASE",'http://www.ecmwf.int')\
         .add_variable("ECF_URL",'"publications/manuals/sms"')\
         .add_variable(a_dict)
    print(suite)
    assert len(list(suite.variables)) == 8,"Expected 8 variable"    
    suite.delete_variable("");         assert len(list(suite.variables)) == 0,"Expected 0 variable since we should have deleted all"
    
    
    # add and delete limits
    the_limit = Limit("limitName1", 10)
    assert len(list(the_limit.node_paths())) == 0 ,"Expected nodes which have consumed a limit to be empty"
    suite.add_limit(  Limit("limitName1", 10) )\
        .add_limit(  Limit("limitName2", 10) )\
        .add_limit(  "limitName3", 10 )\
        .add_limit(  "limitName4", 10 )
    assert len(list(suite.limits)) == 4,"Expected 4 Limits"
    suite.delete_limit("");           assert len(list(suite.limits)) == 0,"Expected 0 limits since we just deleted all of them"


    # add and delete in-limits
    suite.add_inlimit( InLimit("limitName1","/s1/f1",2,True) )\
        .add_inlimit( InLimit("limitName2","/s1/f1",2))\
        .add_inlimit( "limitName3","/s1/f1",2,True)\
        .add_inlimit( "limitName4","/s1/f1",2)
    assert len(list(suite.inlimits)) == 4,"Expected 4 inLimits"
    suite.delete_inlimit("limitName1"); assert len(list(suite.inlimits)) == 3,"Expected 3 inlimits since we just deleted one limitName1" 
    suite.delete_inlimit("");           assert len(list(suite.inlimits)) == 0,"Expected 0 inlimits since we just deleted all of them"


    # add and delete triggers and complete
    task = Task("task")
    task.add_trigger( "t2 == active" ).add_complete( "t2 == complete" )
    assert task.get_complete(), "Expected complete"
    assert task.get_trigger(),  "Expected trigger"
    task.delete_trigger();  assert not task.get_trigger(),  "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"
    
    task.add_part_trigger(  PartExpression("t1 == complete") )\
        .add_part_complete( PartExpression("t3 == complete") ) \
        .add_part_complete( PartExpression("t4 == active",False) )   
    task.delete_trigger();  assert not task.get_trigger(),  "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"

    task.add_part_trigger(  "t1 == complete" ) \
        .add_part_complete( "t3 == complete" ) \
        .add_part_complete( "t4 == active",False)  #  for long and/or expressions, subsequent expr must be and/or
    task.delete_trigger();  assert not task.get_trigger(),  "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"
    
    
    # add and delete events
    task.add_event( Event(1) ).add_event( 2 ).add_event( Event(10, "Eventname") ).add_event( 10, "Eventname2" ).add_event( "fred" )
    assert len(list(task.events)) == 5,"Expected 5 Events"
    task.delete_event("");          assert len(list(task.events)) == 0,"Expected 0 Events"


    # add and delete meter
    task.add_meter( Meter("metername1", 0, 100, 50) )\
        .add_meter( Meter("metername2", 0, 100) )\
        .add_meter( "metername3", 0, 100, 50 )\
        .add_meter( "metername4", 0, 100 )
    assert len(list(task.meters)) == 4,"Expected 4 Meters"
    task.delete_meter("");           assert len(list(task.meters)) == 0,"Expected 0 Meters"

    # add and delete queue
    queue_items = [ "001", "002"]
    task.add_queue( Queue("q1", queue_items) )\
        .add_queue( Queue("q2", queue_items) )\
        .add_queue( "q3", queue_items )\
        .add_queue( "q4", queue_items)
    assert len(list(task.queues)) == 4,"Expected 4 Queues"
    task.delete_queue("");           assert len(list(task.queues)) == 0,"Expected 0 Queues"

    # add and delete label
    task.add_label( Label("label_name1", "value") )\
        .add_label( Label("label_name2", "value") )\
        .add_label( "label_name3", "value" )\
        .add_label( "label_name4", "value" )
    assert len(list(task.labels)) == 4,"Expected 4 labels"
    task.delete_label("");            assert len(list(task.labels)) == 0,"Expected 0 Labels"

    # add delete Repeat
    task.add_repeat( RepeatInteger("integer", 0, 100, 2) ).add_variable("fred","j")
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    task.add_repeat( RepeatEnumerated("enum", ["red", "green", "blue" ]) ).add_variable("Q","j")
    print(task)
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat( RepeatDate("date", 20100111, 20100115,2) ).add_variable("W","j")
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat(RepeatDateTime("datetime", "20100111T000000", "20100115T000000", "48:00:00")).add_variable("W", "j")
    task.delete_repeat()
    repeat = task.get_repeat();
    assert repeat.empty(), "Expected no repeat"

    task.add_repeat( RepeatDateList("date",[20100111, 20100115]) ).add_variable("W","j")
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    task.add_repeat( RepeatString("string", ["a", "b", "c" ] ) ).add_variable("E","j")
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    
    # create a time series, used for adding time and today
    start = TimeSlot(0, 0)
    finish = TimeSlot(23, 0)
    incr = TimeSlot(0, 30)
    time_series = TimeSeries( start, finish, incr, True)

    # Add and delete today
    # NOTE: **********************************************************************
    # Do *NOT* delete attributes using iterator traversal
    # The iterators *are* bound to the c++ iterators hence if we delete
    # over the traversal, we can corrupt the vector, leading to undefined behaviour.  
    # Hence we **cannot** delete more than once over a traversal
    # Soln 1: take a copy
    # *****************************************************************************
    # task.add_today( Today( 0,10 ))
    # task.add_today( Today( 0,59, True ))
    # task.add_today( Today(TimeSlot(20,10)) )
    # task.add_today( Today(TimeSlot(20,20),False)) 
    # for today in task.todays: 
    #    task.delete_today(today) # will corrupt C++ vector
  
    task.add_today( "00:30" ).add_today( "+00:30" ).add_today( "+00:30 20:00 01:00" )\
        .add_today( Today( time_series) )\
        .add_today( Today( 0,10 )).add_today( 0, 59, True )\
        .add_today( Today(TimeSlot(20,10)) )\
        .add_today( Today(TimeSlot(20,20),False)) 
    assert len(list(task.todays)) == 8,"Expected 8 todays"
    vec_copy = []
    for today in task.todays: vec_copy.append(today)
    for today in vec_copy: task.delete_today(today)
    assert len(list(task.todays)) == 0,"Expected 0 todays"
        
             
    # add and delete time
    task.add_time( "00:30" ).add_time( "+00:30" ).add_time( "+00:30 20:00 01:00" )\
        .add_time( Time(time_series ))\
        .add_time( Time( 0,10 ))\
        .add_time( 0, 59, True)\
        .add_time( Time(TimeSlot(20,10)) )\
        .add_time( Time(TimeSlot(20,20),False)) 
    assert len(list(task.times)) == 8,"Expected 8 times"
    vec_copy = []
    for time in task.times: vec_copy.append(time)
    for time in vec_copy: task.delete_time(time)
    assert len(list(task.todays)) == 0,"Expected 0 todays"


    # add and delete date
    for i in [ 1, 2, 4, 8, 16 ] :
        task.add_date( i, 0, 0)
    task.add_date( Date(1, 1, 2010))\
        .add_date( Date(2, 1, 2010))\
        .add_date( Date(3, 1, 2010))\
        .add_date( Date(4, 1, 2010))\
        .add_date( 1, 1, 2010)
    assert len(list(task.dates)) == 10,"Expected 10 dates but found " + str(len(list(task.dates)))
    vec_copy = []
    for attr in task.dates: vec_copy.append(attr)
    for attr in vec_copy: task.delete_date(attr)
    assert len(list(task.dates)) == 0,"Expected 0 dates"

    # add and delete day
    task.add_day( Day(Days.sunday))\
        .add_day( Days.monday)\
        .add_day( Days.tuesday)\
        .add_day( "sunday")     
    assert len(list(task.days)) == 4,"Expected 4 days"
    vec_copy = []
    for attr in task.days: vec_copy.append(attr)
    for attr in vec_copy: task.delete_day(attr)
    assert len(list(task.days)) == 0,"Expected 0 days"
    
    # add and delete crons
    cron = Cron()
    cron.set_week_days( [0, 1, 2, 3, 4, 5, 6]  )
    cron.set_days_of_month( [1, 2, 3, 4, 5, 6 ] )
    cron.set_months(  [1, 2, 3, 4, 5, 6]  )
    start = TimeSlot(0 ,0)
    finish = TimeSlot(23, 0)
    incr = TimeSlot(0, 30)
    ts = TimeSeries( start, finish, incr, True)  # True means relative to suite start
    cron.set_time_series( ts )

    cron1 = Cron()
    cron1.set_week_days( [0, 1, 2, 3, 4, 5, 6 ] )
    cron1.set_time_series( 1, 30,  True )
    
    cron2 = Cron()
    cron2.set_week_days( [0, 1, 2, 3, 4, 5, 6] )
    cron2.set_time_series( "00:30 01:30 00:01" )
    
    cron3 = Cron()
    cron3.set_week_days( [0, 1, 2, 3, 4, 5, 6] )
    cron3.set_time_series( "+00:30" )

    task.add_cron( cron ) \
        .add_cron( cron1 ) \
        .add_cron( cron2 ) \
        .add_cron( cron3 )
    assert len(list(task.crons)) == 4,"Expected 4 crons"
    vec_copy = []
    for attr in task.crons: vec_copy.append(attr)
    for attr in vec_copy: task.delete_cron(attr)
    assert len(list(task.crons)) == 0,"Expected 0 crons"  
    
    
    # add autocancel
    t1 = Task("t1")
    t1.add_autocancel( 3 ).add_variable("A","j")                       
    t3 = Task("t3")
    t3.add_autocancel( 20, 10, True ).add_variable("B","j") 
    t4 = Task("t4")
    t4.add_autocancel( TimeSlot(10, 10), True ).add_variable("C","j") 
    t5 = Task("t5")
    t5.add_autocancel( Autocancel(1, 10, True) ).add_variable("D","j") 
    
    
    # add late
    late = Late()
    late.submitted( TimeSlot(20, 10) )
    late.active(    TimeSlot(20, 10))
    late.complete(  TimeSlot(20, 10), True)
    task.add_late(late ).add_variable("FRED33","j") 
        
    
    # add defstatus, last one set takes effect
    task.add_defstatus( DState.complete )\
        .add_defstatus( DState.queued )\
        .add_defstatus( DState.aborted )\
        .add_defstatus( DState.submitted )\
        .add_defstatus( DState.suspended )\
        .add_defstatus( DState.active )
    
    #
    # add clock
    #
    clock = Clock(1, 1, 2010, False)     # day,month, year, hybrid
    clock.set_gain(1, 10, True)         # True means positive gain
    suite =  Suite("suite")
    suite.add_clock(clock).add_variable("fred1","j") 
    
    end_clock = Clock(1, 1, 2017, False)     # day,month, year, hybrid
    suite.add_end_clock(end_clock).add_variable("fred2","j") 
    
    clock = Clock(1, 1, 2011, True)       # day,month, year, hybrid
    clock.set_gain_in_seconds(12, True)
    s1 =  Suite("s1")
    s1.add_clock(clock).add_variable("fred3","j") 
    s1.add_end_clock(end_clock).add_variable("fred4","j") 
    
    #
    # Add zombie. Note we can *NOT* add two zombie attributes of the same ZombieType
    #
    zombie_life_time_in_server = 800
    child_list = [ ChildCmdType.init, ChildCmdType.event, ChildCmdType.meter, ChildCmdType.label, ChildCmdType.wait, ChildCmdType.abort, ChildCmdType.complete ]
    zombie_type_list = [ ZombieType.ecf, ZombieType.user, ZombieType.path ]
    count = 1;
    for zombie_type in zombie_type_list:
        zombie_attr = ZombieAttr(zombie_type, child_list, ZombieUserActionType.block, zombie_life_time_in_server)
        s1.add_zombie(zombie_attr).add_variable("afred" + str(count),"j") 
        count += 1
    assert len(list(s1.zombies)) == 3,"Expected 3 zombie attributes but found " + str(len(list(s1.zombies)))
    
    # delete all the zombies
    s1.delete_zombie("")
    assert len(list(s1.zombies)) == 0,"Expected zero zombie attributes but found " + str(len(list(s1.zombies)))

    # repeat the the test with empty child list. Empty child list means apply to all child commands
    # + don't specify zombie_life_time_in_server as this is optional
    child_list = [ ]
    zombie_type_list = [ ZombieType.ecf, ZombieType.user, ZombieType.path ]
    for zombie_type in zombie_type_list:
        zombie_attr = ZombieAttr(zombie_type, child_list, ZombieUserActionType.block)
        s1.add_zombie(zombie_attr).add_variable("bfred" + str(count),"j") 
        
        # test delete of specific zombie attribute
        assert len(list(s1.zombies)) == 1,"Expected 1 zombie attributes but found " + str(len(list(s1.zombies)))
        s1.delete_zombie(zombie_type)
        assert len(list(s1.zombies)) == 0,"Expected 0 zombie attributes but found " + str(len(list(s1.zombies)))
        count += 1;

    print("All Tests pass")



# find all cpp file that have std::runtime_error.
# if these file dont have #include <stdexcept> add it as the first include

import os,fnmatch

def all_files(root, patterns='*', single_level=False, yield_folders=False):
    """Expand patterns from semi-colon separated string to list"""
    patterns = patterns.split(';')
    for path, subdirs, files in os.walk(root):
        if yield_folders:
            files.extend(subdirs)
        files.sort()
        for name in files:
            for pattern in patterns:
                if fnmatch.fnmatch(name,pattern):
                    yield os.path.join(path, name)
                    break
        if single_level:
            break    
        
for cpp_file in all_files('.', '*.cpp'):
    if 'cereal' in cpp_file:
        continue
    #print(cpp_file)
    file_obj = open(cpp_file,'r')
    try:
        list_of_all_lines = file_obj.readlines()
        
        has_runtime_error = False
        has_correct_include = False
        for line in list_of_all_lines:
            if 'std::runtime_error' in line:
                has_runtime_error = True
            if '#include <stdexcept>' in line:
                has_correct_include = True
                
        if has_runtime_error:
            #print(cpp_file, " uses std::runtime_error ", has_correct_include )
            if not has_correct_include:
                print(cpp_file," modifying")
                # add #include <stdexcept>
                added_include = False
                new_list = []
                for line in list_of_all_lines:
                    if not added_include and '#include' in line:
                        new_list.append('#include <stdexcept>\n')
                        added_include = True
                    new_list.append(line)
                    
                changed_file = open(cpp_file,'w')
                changed_file.writelines(new_list)
                changed_file.close()
    finally:
        file_obj.close()
        


#!/usr/bin/env python
# This software is provided under the ECMWF standard software license agreement.
""" a layer over raw ecflow api """

# file deepcode ignore PythonConstantBinaryExpression: <comment the reason here>
import unittest

from ecflow import JobCreationCtrl as JobCreationCtrl
# from ecflow import Client as Client
from ecflow import ChildCmdType, TimeSlot
from ecflow import ZombieType, ZombieAttr, ZombieUserActionType
import ecflow, argparse

DEBUG = 0
DECORATE = "ALL"
# DECORATE = "ONLY_TRIGGER"
# DECORATE = "NO_TRIGGER"
# DECORATE = "ONLY_EVENT"
# DECORATE = "NO_ATTRIBUTE"

USE_TIME = True
USE_LATE = False
USE_TRIGGER = True
USE_LIMIT = True
ECF_MODE = "ecflow"
# ECF_MODE = "sms"

if DECORATE == "ONLY_TRIGGER":
    USE_TRIGGER = True 
    USE_LIMIT = False 
    USE_EVENT = False
elif DECORATE == "NO_ATTRIBUTE":
    USE_TRIGGER = False
    USE_LIMIT = False
    USE_EVENT = False
elif DECORATE == "ONLY_EVENT":
    USE_TRIGGER = False
    USE_LIMIT = False
    USE_EVENT = True
elif DECORATE == "NO_TRIGGER":
    USE_TRIGGER = False
    USE_LIMIT = True
    USE_EVENT = True
elif DECORATE == "ALL":
    USE_TRIGGER = True
    USE_LIMIT = True
    USE_EVENT = True
else:
    raise BaseException


def translate(name, value=None):
    """ Translate from sms to ecflow """
    import sms2ecf
    if ECF_MODE != "sms":
        return name, value
    else: sms2ecf.ECF_MODE = ECF_MODE
    return sms2ecf.translate(name, value)

class State:
    """ this class aims at affording a user the possibility to add Triggers as
    t1 = Task("t1")
    Task("ts").add(Trigger(t1 == COMPLETE))

    SUBMITTED, ACTIVE, SUSPENDED, ABORTED, QUEUED, COMPLETE, UNKNOWN
    are instance of this class.
    """

    def __init__(self, state):
        """ store the status """
        self.state = str(state)
    
    def __str__(self):
        """ translate into string """
        return "%s" % self.state

    def __eq__(self, arg):
        """ when == is used, we should care about task name starting with 0-9"""
        if type(arg) == str:
            add = ""
            if type(arg[0]) == int:
                add = "./"                
            return add + arg + " == " + self.state
        elif isinstance(arg, ecflow.Node):
            return arg.get_abs_node_path() + " == " + self.state
        return False

    def __ne__(self, arg):
        """ aka != """
        if type(arg) == str:
            add = ""
            if type(arg[0]) == int:
                add = "./"                
            return add + arg + " != " + self.state
        elif isinstance(arg, ecflow.Node):
            return arg.get_abs_node_path() + " != " + self.state
        return False

    def value(self):
        """ return state """
        return self.state

    def eval(self, node):
        """ return state """
        return self.state == node.get_state()



SUBMITTED = State("submitted")
ACTIVE    = State("active")
SUSPENDED = State("suspended")
ABORTED   = State("aborted")
QUEUED    = State("queued")
COMPLETE  = State("complete")
UNKNOWN   = State("unknown")

class Attribute(object):
    """ generic attribute to be attached to a node"""
    def add_to(self, node):
        """ use polymorphism to attach attribute to a node"""
        pass

class Variable(Attribute, ecflow.Variable):
    """ filter variables to be attached to a node, may alter its name SMS-ECF"""
    def add_to(self, node):
        """ add_variable"""
        keyt, valt = translate(self.name(), self.value())
        node.add_variable(keyt, valt)

class Label(Attribute, ecflow.Label):
    """ wrap around label"""
    def add_to(self, node):
        """ add_label"""
        node.add_label(self)

class Meter(Attribute, ecflow.Meter):
    """ wrap around meter"""
    def add_to(self, node):
        """ add_meter"""
        node.add_meter(self)

class Event(Attribute, ecflow.Event):
    """ wrap around event"""
    def add_to(self, node):
        """ add_event"""
        node.add_event(self)

class InLimit(Attribute):
    """ a class to host a path for a limit
        silently ignore if USE_LIMIT is False, 
        (in debug mode) """

    def __init__(self, fullpath):
        self.data = None
        if USE_LIMIT: 
            try: path, name = fullpath.split(":")
            except: name = fullpath; path = ""
            if name is None: 
                raise BaseException
            self.data  = ecflow.InLimit(name, path)
            self.path_ = path
            self.name_ = name

    def add_to(self, node):
        """ add_inlimit"""
        if USE_LIMIT and self.data is None: 
            raise BaseException
        if  USE_LIMIT and self.data is not None:
            node.add_inlimit(self.data)

    def value(self):        
        """ get limit fullpath-name """
        return self.path_ + ":" + self.name_

    def name(self):
        """ get limit name """
        return self.name_

class Inlimit(InLimit):
    def __init__(self, fullpath):
        super(Inlimit, self).__init__(fullpath)

class Trigger(Attribute):
    """ add trigger (string, list of task names, or directly
           expression and: and'ed (True) or or'ed (False) unk: add or
           [name]==unknown for RD
    """
    def __init__(self, expr, unk=False, anded=True):
        self.expr = None
        if expr is None: 
            return
        if expr == "": 
            return
        if type(expr) == str:
            # self.expr = ecflow.Expression(expr)
            self.expr = expr
            return

        if type(expr) == tuple:
            prep = list(expr)
            expr = prep

        if type(expr) == list:
            for index, name in enumerate(expr):
                if name is None: 
                    continue
                pre = ""

                if type(name) in (Node, Task, Family, Suite):
                    fullname = name.fullname()
                    name = fullname
                elif type(name) in (ecflow.Task, ecflow.Family):
                    name = name.name()

                if name[0].isdigit(): # in "0123456789":
                    pre = "./"

                if ':' in name:
                    item = pre + name
                else:
                    item = pre + "%s == complete" % name
                    if unk:
                        item += " or %s%s==unknown" % (pre, name)

                if index == 0:
                    self.expr = ecflow.Expression(item)
                else:
                    self.expr.add(ecflow.PartExpression(item,anded))

        elif type(expr) in (ecflow.Expression, 
                            ecflow.PartExpression):
            self.expr = ecflow.Expression(str(item))
        else:
            print type(expr)
            raise Exception("what? trigger?")

    def add_to(self, node):
        if not USE_TRIGGER:
            return
        if self.expr is None:
            return
        if node.get_trigger() is None:
            node.add_trigger(self.expr)
        else:
            node.add_part_trigger(ecflow.PartExpression(self.expr, True))

class TriggerAnd(Trigger):
    def __init__(self, expr, unk=False, anded=True):
        # super(TriggerAnd, self).__init__(expr, unk, anded)
        self.expr = expr
        if not ":" in expr and not " eq " in expr and not "==" in expr:
            self.expr += " eq complete"
    def add_to(self, node):        
        node.add_part_trigger(ecflow.PartExpression(self.expr, True))

class TriggerImpossible(Trigger):
    """ attribute to be added to node when it is not expected to run
    any task"""

    def __init__(self):
        """ add an 'impossible trigger', for a task not to run """
        super(TriggerImpossible, self).__init__("1==0")

class TriggerAlways(Trigger):
    """ attribute to be added to node when it is not expected to run
    any task"""

    def __init__(self):
        """ add an 'impossible trigger', for a task not to run """
        super(TriggerAlways, self).__init__("1==1")

class Complete(Trigger):
    """ class to host complete expression, added later to a node"""
    def __init__(self, expression, unk=False, anded=False):
        super(Complete, self).__init__(expression, unk, anded)

    def add_to(self, node):
        if USE_TRIGGER and self.expr is not None:
            node.add_complete(self.expr)

class Clock(Attribute):
    """ wrapper to add clock """
    def __init__(self, arg):
        if type(arg) == str:
            hybrid = "hybrid" in arg
            hhh, mmm, sss = [0, 0, 0]
            try:
                hhh, mmm, sss = arg.split(':')
                self.data = ecflow.Clock(hhh, mmm, sss, hybrid)
            except:
                self.data = ecflow.Clock(hybrid)
        else:
            self.data = ecflow.Clock(arg)

    def add_to(self, node):
        if type(node) != Suite:
            print "WAR: clock can only be attached to suite node, "
            print "WAR: clock is ignored"
            return
        node.add_clock(self.data)

class AutoCancel(Attribute):
    """ wrapper to add time """
    def __init__(self, arg):
        if type(arg) == str:
            hhh, mmm = arg.split(':')
            rel = '+' in arg
            self.data = ecflow.Autocancel(int(hhh), int(mmm), rel)
        else:
            self.data = ecflow.Autocancel(arg)

    def add_to(self, node):
        node.add_autocancel(self.data)

class Time(Attribute):
    """ wrapper to add time """
    def __init__(self, arg):
        self.data = arg

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.add_time(self.data)

class Today(Time):
    """ wrapper to add time """
    def __init__(self, arg):
        self.data = arg

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.add_today(self.data)

class Cron(Time):
    """ wrapper to add time """
    def __init__(self, bes, wdays=None, days=None, months=None):
        if type(bes) == ecflow.Cron:
            self.data = arg
            return

        # parser = argparse.ArgumentParser()
        # parser.add_argument("-w", default=0, help= "weekdays")
        # parser.add_argument("-d", default=0, help= "days")
        # parser.add_argument("-m", default=0, help= "months")
        # parser.add_argument("bes", nargs='*', type=str, help= "begin end step")
        # parsed = parser.parse_args(arg)

        # self.data = ecflow.Cron()
        # if parsed.w: self.data.set_week_days(parsed.w)
        # if parsed.d: self.data.set_days_of_month(parsed.d)
        # if parsed.m: self.data.set_months(parsed.m)
        # cli = ' '.join(''.join(parsed.bes).split(' '))
        # print cli, parsed.bes, parsed.w, parsed.d, parsed.m
        # self.data.set_time_series(cli)
        self.data = ecflow.Cron()
        if wdays: 
            self.data.set_week_days([int(x) for x in wdays.split(',')])
        if days: 
            self.data.set_days_of_month([int(x) for x in days.split(',')])
        if months: 
            self.data.set_months([int(x) for x in months.split(',')])
        self.data.set_time_series(bes)
    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.add_cron(self.data)

class Date(Time):
    """ wrapper to add date """

    def __init__(self, arg, mask=False):
        super(Date, self).__init__(arg)
        self.mask = mask

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            ### ??? FIX, emos avoids dates, datasvc would not
            if self.mask:
                node.add_variable("DATEMASK", self.data)
            else:
                ddd, mmm, yyy = self.data.rsplit('.')
                if ddd == '*': ddd = 0
                if mmm == '*': mmm = 0
                if yyy == '*': yyy = 0

                node.add_date(int(ddd), int(mmm), int(yyy))
                # node.add_date(self.data)

class Day(Date):
    """ wrapper to add day """
    def add_to(self, node):
        if USE_TIME and self.data is not None:
            # node.add_date(self.data) ### FIX
            node.add_variable("WEEKDAY", self.data)

class Defcomplete(Attribute):
    """ wrapper to add defstatus complete """
    def __init__(self):
        pass

    def add_to(self, node):
        node.defstatus("complete")

class Defstatus(Defcomplete):
    """ add defstatus attribute"""
    def __init__(self, kind):
        if type(kind) == str:
            kinds = {"suspended": ecflow.DState.suspended,
                     "aborted": ecflow.DState.aborted,
                     "complete": ecflow.DState.complete,
                     "active": ecflow.DState.active,
                     "submitted": ecflow.DState.submitted,
                     "unknown": ecflow.DState.unknown,
                     "queued": ecflow.DState.queued,
                     }
            self.data = kinds[kind]
        else:
            self.data = kind

    def add_to(self, node):
        node.add_defstatus(self.data)

class DefcompleteIf(Defcomplete):
    """ wrapper to add conditional defstatus complete 
    just change name to make it explicit
    """
    def __init__(self, arg=True):
        # super(DefcompleteIf, self).__init__()
        self.data = arg

    def add_to(self, node):
        if self.data:
            node.defstatus("complete")
        # else: node.defstatus("queued") # in comment to prevent
        # overwrite when using multiple defcomplete

class Limit(Attribute):
    """ wrapper to add limit """

    name = None
    size = 1

    def __init__(self, nam=None, lim=1):
        self.name = nam
        self.size = lim

    def add_to(self, node):
        if USE_LIMIT and self.name is not None:
            node.add_limit(self.name, self.size)

class Late(Attribute):
    """ wrapper around late, to be add'ed to families and tasks """

    def __init__(self, arg):
        self.data = None
        # if not USE_LATE: if 1: raise
        # return
        sub = False
        act = False
        com = False
        rel = False
        self.data = ecflow.Late()
        for item in arg.split(" "):
            if   item == "-s": 
                sub = True
            elif item == "-c": 
                com = True
            elif item == "-a": 
                act = True
            else:
                hour, mins = item.split(":")
                rel = "+" in hour
                if   sub: 
                    self._add_sub(hour, mins)
                elif com: 
                    self._add_com(hour, mins, rel)
                elif act: 
                    self._add_act(hour, mins, rel)
                sub = False
                act = False
                com = False

    def _add_sub(self, hour, mins):
        """ submitted"""
        self.data.submitted(int(hour), int(mins))

    def _add_com(self, hour, mins, rel):
        """ complete"""
        self.data.complete(ecflow.TimeSlot(int(hour), int(mins)), rel)

    def _add_act(self, hour, mins, rel):
        """ active"""
        self.data.active(ecflow.TimeSlot(int(hour), int(mins)), rel)

    def add_to(self, node):
        if USE_LATE and self.data is not None:
            node.add_late(self.data)

class Variables(Attribute):
    """ dedicated class to enable variable addition with different
    syntax """
    
    def _set_tvar(self, key, val):        
        """ facilitate to load an ecflow suite to SMS, translating
        variable names"""
        keyt, valt = translate(str(key), str(val))
        if self.data is None:
            self.data = Variable(keyt, valt)
        else:
            next = self.next
            self.next = Variables(keyt, valt, next)

    def __init__(self, __a=None, __b=None, __next=None, *args, **kwargs):
        self.data = None
        self.next = __next

        if len(args) > 0:
            if type(args) == list:
                for item in args.iteritems():
                    self._set_tvar(item.name(), item.value())
            elif type(args) == tuple:
                for key, val in args.items():
                    self._set_tvar(key, val)
            else:
                raise BaseException()
        if len(kwargs) > 0: 
            for key, val in kwargs.items():
                self._set_tvar(key, val)
        if type(__a) == dict:
            for key, val in __a.items():
                self._set_tvar(key, val)
        elif type(__a) == tuple: 
            raise BaseException()
            # for key, val in __a.items(): self._set_tvar(key, val)
        elif type(__a) == list: raise BaseException()
        elif type(__a) == Variable: self.data = __a
        elif __a is not None and __b is not None:
            self._set_tvar(__a, __b)
        elif __a is None and __b is None: pass
        else:
            print __a, __b, __next, args, kwargs; 
            raise BaseException()

    def add_to(self, node):
        if self.data is not None:
            node.add_variable(self.data)
        if self.next is not None:
            self.next.add_to(node)

    def add(self, what):
        print what.fullname()
        raise baseException

class Limits(Attribute):
    """ dedicated class to enable limits addition with different syntax """
    
    def _set_tvar(self, key, val):        
        """ append limits """
        if self.data is None:
            self.data = Limit(key, val)
        else:
            next = self.next
            self.next = Limits(key, val, next)

    def __init__(self, __a=None, __b=None, __next=None, *args, **kwargs):
        self.data = None
        self.next = __next

        if len(args) > 0:
            if type(args) == list:
                for item in args.iteritems():
                    self._set_tvar(item.name(), item.value())
            elif type(args) == tuple:
                for key, val in args.items():
                    self._set_tvar(key, val)
        elif len(kwargs) > 0: 
            for key, val in kwargs.items():
                self._set_tvar(key, val)
        elif type(__a) == dict:
            for key, val in __a.items():
                self._set_tvar(key, val)

        if __a is not None and __b is not None:
            self._set_tvar(__a, __b)

    def add_to(self, node):
        if self.data is not None:
            node.add_limit(self.data)
        if self.next is not None:
            self.next.add_to(node)

class Repeat(Attribute):
    def __init__(self, name="YMD", start=20120101, end=20120101, step=1, kind="date"):
       if kind == "date":
           self.data = ecflow.RepeatDate(name, int(start), int(end), 
                                         int(step))
       elif kind == "integer":
           self.data = ecflow.RepeatInteger(name, int(start), int(end), int(step))
       elif kind == "string":
           self.data = ecflow.RepeatString(name, start)
       elif kind == "enumerated":
           self.data = ecflow.RepeatEnumerated(name, start)
       elif kind == "day":
           self.data = ecflow.RepeatDay(step)        
       else:
           self.data = None

    def add_to(self, node):
        if self.data is not None:
            node.add_repeat(self.data)

def If(test=True, then=None, otow=None):
    """ enable Task("t1").add(If(test= (1==1),
                                 then= Variables(ONE=1),
                                 otow= Variables(TWO=2)))
        appreciate that both branches are evaluated, using this If class
        ie there is no 'dead code' as it is with python language 'if' structure

        using If to distinguish od/rd mode request that both users share
        the variables (parameter.py) and ecf.py
                                 
        otow: on the other way?
        """
    if test:
        return then
    return otow

class Root(object): # from where Suite and Node derive
    """ generic tree node """

    def __str__(self):
        if isinstance(self, ecflow.Node):
            return self.fullname()
        return str(self)

    def __eq__(self, node):
        if isinstance(self, ecflow.Node):
            return "%s == " % self.fullname() + str(node)
        return False

    def __ne__(self, node):        
        if isinstance(self, ecflow.Node):
            return "%s != " % self.fullname() + str(node)
        return False

    def __and__(self, node):
        if isinstance(self, ecflow.Node):
            return "%s and " % self.fullname() + str(node)
        return False

    def __or__(self, node):
        if isinstance(self, ecflow.Node):
            return "%s or " % self.fullname() + str(node)
        return False

    def fullname(self):
        """ simple syntax """
        if isinstance(self, ecflow.Node):
            return self.get_abs_node_path()
        return str(self)
    
    def repeat(self, name="YMD", start=20120101, end=20321212, step=1,
               kind="date"):
        """ add repeat attribute"""
        if kind == "date":
            self.add_repeat(ecflow.RepeatDate(name, int(start), int(end), 
                                              int(step)))
        elif kind == "integer":
            self.add_repeat(ecflow.RepeatInteger(name, int(start), int(end), int(step)))
        elif kind == "string":
            self.add_repeat(ecflow.RepeatString(name, start))
        elif kind == "enumerated":
            self.add_repeat(ecflow.RepeatEnumerated(name, start))
        elif kind == "day":
            self.add_repeat(ecflow.RepeatDay(step))
        else:
            raise BaseException
        return self

    def defstatus(self, kind):
        """ add defstatus attribute"""
        status = kind
        if type(kind) == str:
            kinds = {"suspended": ecflow.DState.suspended,
                     "aborted": ecflow.DState.aborted,
                     "complete": ecflow.DState.complete,
                     "active": ecflow.DState.active,
                     "submitted": ecflow.DState.submitted,
                     "unknown": ecflow.DState.unknown,
                     "queued": ecflow.DState.queued,
                     }
            status = kinds[kind]
        self.add_defstatus(status)
        return self

    def add(self, item=None, *args):
        """ add a task, a family or an attribute """
        if DEBUG: print self.fullname()

        if item is not None:
            if type(item) == tuple:
                for val in item:
                    self.add(val)
            elif type(item) == list:
                for val in item:
                    self.add(val)
            else:
                try:
                    item.add_to(self)
                except Exception, exc:
                    try: print item.fullname(), item.name(), self.name()
                    except: pass
                    print "not yet", self, type(item), item, exc 
                    raise BaseException

        if len(args) > 0:
            if type(args) == tuple:
                for val in args:
                    self.add(val)
            elif type(args) == list:
                for val in args:
                    self.add(val)
            else:
                raise BaseException()

        if not isinstance(self, ecflow.Node):
            print "you don't want that"
            raise BaseException()

        return self
 
    def limit(self, name, size):
        """ add limit attribute"""
        if name is None: 
            raise BaseException
        self.add_limit(name, size)
        return self

    def inlimit(self, full_path):
        """ add inlimit attribute"""
        if not USE_LIMIT: 
            return self

        path, name = full_path.split(":")
        if name is None:
            raise BaseException
        if path is None:
            raise BaseException

        self.add_inlimit(name, path)
        return self

class Node(Root): # from where Task and Family derive
    """ Node class is shared by family and task """

    def add_limits(self, __a = None, __b = None, **kwargs):
        """ add limit dependency"""
        if isinstance(__a, basestring):
            self.add_limit(__a, __b)
        elif isinstance(__a, dict):
            assert __b is None
            for key, val in __a.items():
                self.add_limit(key, val)
        for key, val in kwargs.items():
            self.add_limit(key, val)
        return self

    def meter(self, name, start, end, threshold=None):
        """ add meter attribute"""
        if threshold is None:
            threshold = end
        self.add_meter(name, start, end, threshold)
        return self

    def label(self, name, default=""):
        """ add label attribute"""
        self.add_label(name, default)
        return self

    def event(self, name=1):
        """ add event attribute"""
        if USE_EVENT: 
            self.add_event(name)
        return self

    def cron(self, time, dom=False, wdays=False, month=False):
        """ wrapper for add_cron """
        cron = ecflow.Cron()
        cron.set_time_series(time)
        if wdays is not False:
            cron.set_week_days(wdays)
        if month is not False:
            cron.set_months(month)
        if dom is not False:
            cron.set_day_of_month(dom)

        self.add_cron(cron)
        return self

    def today(self, hhmm): 
        """ wrapper around time """
        self.time(hhmm)
        return self # ???

    def time(self, hhmm):
        """ wrapper around time, None argument is silently ignored """
        if hhmm is not None:
            self.add_time(hhmm)
        return self

    def trigger(self, arg):
        """ add trigger attribute"""
        if USE_TRIGGER and arg is not None:
            self.add_trigger(arg)
        return self

    def trigger_and(self, arg):
        """ append to existing trigger"""
        if USE_TRIGGER and arg is not None:
            self.add_part_trigger(ecflow.PartExpression(arg, True))
        return self

    def trigger_or(self, arg):
        """ append to existing trigger"""
        if USE_TRIGGER and arg is not None:
            self.add_part_trigger(ecflow.PartExpression(arg, False))
        return self

    def complete(self, arg):
        """ add complete attribute"""
        if USE_TRIGGER and arg is not None:
            self.add_complete(arg)
        return self

    def complete_and(self, arg):
        """ append to existing complete"""
        if USE_TRIGGER and arg is not None:
            self.add_part_complete(ecflow.PartExpression(arg, True))
        return self

    def complete_or(self, arg):
        """ append to existing complete"""
        if USE_TRIGGER and arg is not None:
            self.add_part_complete(ecflow.PartExpression(arg, False))
        return self

    def up(self):
        """ get parent, one level up"""
        return self.get_parent()

class Defs(ecflow.Defs):
    """ wrapper for the definition """
    def add(self, suite):
        """ add suite """
        self.add_suite(suite)
        return suite

    def suite(self, name):
        """ add suite providing its name """
        suite = Suite(name)
        self.add(suite)
        return suite

class Client(ecflow.Client):
    """ wrapper around client """
    def __init__(self, host="localhost", port="31415"):
        if "@" in host:
            host, port = host.split("@")
            super(Client, self).__init__(host, int(port))            
        else:
            super(Client, self).__init__(host, int(port))
        self.host = host
        self.port = port

    def __str__(self):
        return "#MSG: ecflow client %s@%s v%s" % (self.host, self.port, 
                                                 self.version())

class Suite(ecflow.Suite, Root):
    """ wrapper for a suite """

    def family(self, name):
        """ add family """
        fam = Family(name)
        self.add_family(fam)
        return fam

    def task(self, name):
        """ add family """
        tsk = Task(name)
        self.add_task(tsk)
        return tsk

    # def __enter__(self):
    #     return self

    # def __exit__(self, *args):
    #     pass

class Family(ecflow.Family, Node, Attribute):
    """ wrapper around family """

    def family(self, name):
        """ add a family """
        fam = Family(name)
        self.add_family(fam)
        return fam

    def task(self, name):
        """ add a task """
        tsk = Task(name)
        self.add_task(tsk)
        return tsk

    def add_to(self, node):        
        node.add_family(self)

    # def __enter__(self):
    #     return self

    # def __exit__(self, *args):
    #     pass

class Task(ecflow.Task, Node, Attribute):
    """ wrapper around task """

    def __setattr__(self, key, val):
        # assert key.isupper()
        if key.isupper():
            key, val = translate(key, val)
        self.add_variable(key, val)

    def add_to(self, node):
            node.add_task(self)

def display(defs, fname=None):
    """ print defs"""
    if fname is None:
        print defs
    else:
        fop = open(fname, "w")
        print >> fop, defs


class TestEcf(unittest.TestCase):
    """ a test case """
    def test_xxx(self):
        """ a test """

        suite = Suite ("a_suite")
        suite.defstatus("suspended")

        fam = Family("a_family")
        tsk = Task("a_task")

        ft2 = Task("a_fam")
        ft2.add_to(fam)

        tsk.VAR = "VALUE"                         # edit VAR "VALUE"
        tsk.add(Late("-s 00:05 -c 01:00"))
        fam.add(tsk,

                (Task("1"), Task("2")), 
                [Task("11"), Task("12")], 
                Task("111"), Task("211"),

                Task("t2").add(Trigger(tsk == COMPLETE),
                               Time("01:00"))            )

        fam.add(Task("t3").add(
                If(test= (1==1),
                   then= Variables(ADD_ONE=1),
                   otow= Variables(ADD_TWO=1)),

                If(test= (1==0),
                   then= Variables(ADD_ONE=0),
                   otow= Variables(ADD_TWO=0)),
                Trigger(tsk != ABORTED),
                Complete(tsk == COMPLETE))) # longer

        fam.add(
            Task("2t"),
            Task("t4").add(
                Trigger(tsk.name() != COMPLETE)),
            Late("-s 00:05 -c 01:00"),
            Variables(VAR="VALUE"),
            Task("t5").add(Trigger(["t4", "t3", "t2"])),
            Task("t6").add(Trigger("2t" == COMPLETE)),
            Task("t7").add(Trigger("2t eq complete")),
            )

        tsk.add(Limit("a_limit", 10),
                InLimit("a_task:a_limit"),
                Meter("step", -1, 100),
                Label("info", "none"),
                Event(1),
                Event("a"),
                Defcomplete())
        
        tsk.add(Variables({"A": "a", "B": "b"}))
        tsk.add(Variables(D="d", E="e"))
        tsk.add(Variables("C", "c"))
        suite.add(fam)

        fam.family("another").add(DefcompleteIf(True))

        defs = Defs()
        defs.add(suite)
        another = defs.suite("another")
        another.defstatus("suspended")
        another.task("tn")
        afam = another.family("another_fam")
        afam.task("t2n")

        display(defs, fname="test_ecf.tmp")

if __name__ == '__main__':
    unittest.main()
# http://intra.ecmwf.int/metapps/monitoring/sst/sst_ostia.png
# http://intra.ecmwf.int/metapps/monitoring/sst/sstdiff_ostia.png



import os
from ecflow import (
    Defs,
    Suite,
    Family,
    Task,
    Edit,
    Trigger,
    Complete,
    Event,
    Meter,
    Time,
    Day,
    Date,
    Label,
    RepeatString,
    RepeatInteger,
    RepeatDate,
    InLimit,
    Limit,
    Late,
)


def create_family_f6():
    return Family(
        "f6", Edit(SLEEP=120), Task("t1", Late(complete="+00:01"))
    )  # set late flag if task t1 takes longer than a minute


print("Creating suite definition")
home = os.path.join(os.getenv("HOME"), "course")
defs = Defs(Suite("test", Edit(ECF_INCLUDE=home, ECF_HOME=home), create_family_f6()))
print(defs)

print("Checking job creation: .ecf -> .job0")
print(defs.check_job_creation())

print("Checking trigger expressions")
assert len(defs.check()) == 0, defs.check()

print("Saving definition to file 'test.def'")
defs.save_as_defs("test.def")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
#  code for testing creation of defs file in python
import os
import ecflow
import copy
from ecflow import Suite, Family, Task, Defs, Clock, DState, PartExpression, Variable, Limit, InLimit, \
                   Date, Day, Event, Meter, Label, Autocancel, Days, TimeSlot, TimeSeries, Style, State, \
                   RepeatString, RepeatDate, RepeatInteger, RepeatDay, RepeatEnumerated, \
                   Verify, PrintStyle, Time, Today, Late, Cron, Client, debug_build
import ecflow_test_util as Test


if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
    
    defs = Defs()
    s0 = defs.add_suite("s0")
    s1 = defs.add_suite("s1")
    f1 =  s1.add_family("f1")
    t1 = f1.add_task("t1")
    t2 = f1.add_task("t2")
    assert len(list(f1.nodes)) == 2 ,"Expected family to have 2 nodes"
    defs_copy = copy.copy(defs)
     
    
    t2_remove = t2.remove()
    assert len(f1) == 1 ,"Expected family to have 1 nodes"

    t1_remove = t1.remove()
    assert len(f1) == 0 ,"Expected family to have 0 nodes"

    f1_remove = f1.remove()
    assert len(s1) == 0 ,"Expected suite to have no nodes"

    node_vec = defs.get_all_nodes()
    assert len(node_vec) == 2, "Expected 2 suites only, but found " + str(len(node_vec))

    s1_remove = s1.remove()
    node_vec = defs.get_all_nodes()
    assert len(node_vec) == 1, "Expected 1 suites only, but found " + str(len(node_vec))

    s0_remove = s0.remove()
    node_vec = defs.get_all_nodes()
    assert len(node_vec) == 0, "Expected 0 suites only, but found " + str(len(node_vec))

    # =========================================================
    defs2 = Defs();
    defs2.add_suite(s0_remove)
    s = defs2.add_suite(s1_remove) # s1_remove is returned as s
    f =  s.add_family(f1_remove)   # f1_remove is returned as f
    f.add_task(t1_remove)
    f.add_task(t2_remove)
    assert defs2 == defs_copy ,"expected defs to be the same"



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
# Utility code used in the tests.
#
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
from socket import gethostname 
import os,sys,fnmatch
import time
import fcntl
import datetime,time
import shutil   # used to remove directory tree
import platform # for python version info

from ecflow import Client, debug_build, File

def all_files(root, patterns='*', single_level=False, yield_folders=False):
    """Expand patterns from semi-colon separated string to list"""
    patterns = patterns.split(';')
    for path, subdirs, files in os.walk(root):
        if yield_folders:
            files.extend(subdirs)
        files.sort()
        for name in files:
            for pattern in patterns:
                if fnmatch.fnmatch(name,pattern):
                    yield os.path.join(path, name)
                    break
        if single_level:
            break    
        
# Enable to stop data being deleted, and stop server from being terminated
def debugging() : return False

def ecf_home(port): 
    # debug_build() is defined for ecflow. Used in test to distinguish debug/release ecflow
    # Vary ECF_HOME based on debug/release/port/python version, allowing multiple invocations of these tests
    ecf_home_base = "/test/data/ecf_home_py" + str(sys.version_info[0])
    if debug_build():
        return os.getcwd() + ecf_home_base + "_debug_" + str(port)
    return os.getcwd() + ecf_home_base + "_release_" + str(port)

def get_parent_dir(file_path):
    return os.path.dirname(file_path)

def log_file_path(port): return "./" + gethostname() + "." + port + ".ecf.log"
def checkpt_file_path(port): return "./" + gethostname() + "." + port + ".ecf.check"
def backup_checkpt_file_path(port): return "./" + gethostname() + "." + port + ".ecf.check.b"
def white_list_file_path(port): return "./" + gethostname() + "." + port + ".ecf.lists"

def print_test_start(test_name):
    print("#######################################################################################")
    print(test_name)  
    print("ecflow version(" + Client().version() + ") debug build(" + str(debug_build()) +")  python(" + platform.python_version() + ")")
    print("PYTHONPATH                : " + str(os.environ['PYTHONPATH'].split(os.pathsep)))
    #print("sys.path                  : " + str(sys.path))
    print("Current working directory : " + str(os.getcwd()))
    print("Python version            : " + str(sys.version_info[0]) + "." + str(sys.version_info[1]))
    print("#######################################################################################")
    
def clean_up_server(port):
    print("   clean_up " + port)
    try: os.remove(log_file_path(port))
    except: pass
    try: os.remove(checkpt_file_path(port))
    except: pass
    try: os.remove(backup_checkpt_file_path(port))
    except: pass
    try: os.remove(white_list_file_path(port))  
    except: pass
    
def clean_up_data(port):
    print("   Attempting to Removing ECF_HOME " + ecf_home(port))
    try: 
        shutil.rmtree(ecf_home(port),True)   # True means ignore errors 
        print("   Remove OK") 
    except: 
        print("   Remove Failed") 
        pass
        
# =======================================================================================
class EcfPortLock(object):
    """allow debug and release version of python tests to run at the same
    time, buy generating a unique port each time"""
    def __init__(self):
        print("   EcfPortLock:__init__")
        pass
    
    def at_time(self):
        return datetime.datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')
    
    def find_free_port(self,seed_port):
        print("   EcfPortLock:find_free_port starting with " + str(seed_port) + " at time " + self.at_time())
        port = seed_port
        while 1:
            # port must be free for at least 15 seconds
            if self._timed_free_port(port,15) == True:
                print("   *FOUND* free server port " + str(port) + " : " + self.at_time())
                if self.do_lock(port):
                    break
            else:
                print("   *Server* port " + str(port) + " busy( by  ping), trying next port " + self.at_time())
            port = port + 1
            
        return str(port)  
    
    def _timed_free_port(self,port,wait_time = 10):
        count = 0
        while count < wait_time:
            if self._free_port(port) == True:
                count = count + 1
                time.sleep(1)
            else:
                return False
        return self._free_port(port)
            
    def _free_port(self,port):
        try:
            ci = Client()
            ci.set_host_port("localhost",str(port))
            ci.ping() 
            return False
        except RuntimeError as e:
            return True
            
    def do_lock(self,port):
        file = self._lock_file(port)
        if os.path.exists(file):
            print("   *LOCKED* lock file exists " + file + " trying next port : " + self.at_time() )
            return False
        try:
            fp = open(file, 'w') 
            try:
                self.lock_time = self.at_time()
                fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
                self.lock_file_fp = fp
                print("   *LOCKED* file " + file + " : " + self.lock_time )
                return True;
            except IOError:
                print("   Could *NOT* lock file " + file + " trying next port : " + self.at_time())
                return False
        except IOError as e:
            print("   Could not open file " + file + " for write trying next port : " + self.at_time())
            return False
        
    def remove(self,port):
        file = self._lock_file(port)
        print("   Remove lock file : " + file + " : lock_time: " + self.lock_time + " release_time: " + self.at_time() )
        self.lock_file_fp.close()
        os.remove(file)
    
    def _lock_file(self,port):
        if "ECF_PORT_LOCK_DIR" in os.environ:
            lock_file = os.environ["ECF_PORT_LOCK_DIR"] + "/" + str(port) + ".lock"
            #print("     EcfPortLock::_lock_file ECF_PORT_LOCK_DIR: " + lock_file)
            return lock_file
        if os.path.exists(File.source_dir()):
            lock_file = File.source_dir() + "/" + str(port) + ".lock"
            #print("     EcfPortLock::_lock_file File::source_dir(): " + lock_file)
            return lock_file
        lock_file = os.getcwd() + "/" + str(port) + ".lock"
        #print("     EcfPortLock::_lock_file os.getcwd(): " + lock_file)
        return lock_file
        
# ===============================================================================

class Server(object):
    """TestServer: allow debug and release version of python tests to run at the same
    time, by generating a unique port each time"""
    def __init__(self):
        print("Server:__init__: Starting server")    
        if not debugging():
            seed_port = int(os.getenv('ECF_FREE_PORT', '3153'))
            if sys.version_info[0] == 3: # python3 can run at same time
                seed_port = int(os.getenv('ECF_FREE_PORT', '3154'))  
            if debug_build(): seed_port = seed_port + 1 
            self.lock_file = EcfPortLock()
            self.the_port = self.lock_file.find_free_port(seed_port)   
        else:
            self.the_port = "3152"
            
        # Only worth doing this test, if the server is running
        # ON HPUX, having only one connection attempt, sometimes fails
        #ci.set_connection_attempts(1)     # improve responsiveness only make 1 attempt to connect to server
        #ci.set_retry_connection_period(0) # Only applicable when make more than one attempt. Added to check api.
        self.ci = Client("localhost", self.the_port)

    def at_time(self):
        return datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')

    def __enter__(self):
        try:
            print("Server:__enter__: About to ping localhost: " + self.the_port +  " : " + self.at_time())       
            self.ci.ping() 
            print("   ------- Server all ready running on port " + self.the_port + " *UNEXPECTED* -- : " + self.at_time() )
            sys.exit(1)
        except RuntimeError as e:
            while 1:
                # Only worth doing this test, if the server is running
                # ON HPUX, having only one connection attempt, sometimes fails
                #ci.set_connection_attempts(1)     # improve responsiveness only make 1 attempt to connect to server
                #ci.set_retry_connection_period(0) # Only applicable when make more than one attempt. Added to check api.
                self.ci = Client("localhost", self.the_port)

                print("   ------- Server *NOT* running on port " + self.the_port + " as *EXPECTED* ------ ") 
                print("   ------- Start the server on port " + self.the_port + " ---------")  
                if not debugging():
                    clean_up_server(str(self.the_port))   
                    clean_up_data(str(self.the_port))    
        
                server_exe = File.find_server();
                assert len(server_exe) != 0, "Could not locate the server executable"
            
                server_exe += " --port=" + self.the_port + " --ecfinterval=4 &"
                print("   TestClient.py: Starting server " + server_exe)
                os.system(server_exe) 
            
                print("   Allow time for server to start " + self.at_time())
                if self.ci.wait_for_server_reply() :
                    print("   Server has started " +  self.at_time())
                    break
                else:
                    print("   Server failed to start after 60 second, trying next port !!!!!! " + self.at_time())
                    self.the_port = self.lock_file.find_free_port( int(self.the_port) + 1 )   
             
        print("   Run the tests, leaving Server:__enter__:") 

        # return the Client, that can call to the server
        return self.ci
    
    def __exit__(self,exctype,value,tb):
        print("Server:__exit__: Kill the server, clean up log file, check pt files and lock files, ECF_HOME " + self.at_time())
        print("   exctype:",exctype)
        print("   value:",value)
        print("   tb:",tb)
        print("   Terminate server "+ self.at_time() )
        self.ci.terminate_server()  
        print("   Terminate server OK " + self.at_time() )
        self.lock_file.remove(self.the_port)
        
        # Do not clean up data, if an assert was raised. This allow debug
        if exctype is None:
            if not debugging():
                clean_up_server(str(self.the_port))
                clean_up_data(str(self.the_port))
        return False
